from logging import Logger
from typing import TYPE_CHECKING
from dataclasses import dataclass, field

from rid_lib.ext import Cache

from ..network.resolver import NetworkResolver
from ..config.base import BaseNodeConfig
from ..config.provider import ConfigProvider
from ..network.graph import NetworkGraph
from ..network.event_queue import EventQueue
from ..network.request_handler import RequestHandler
from ..identity import NodeIdentity
from .kobj_queue import KobjQueue

if TYPE_CHECKING:
    from ..effector import Effector


@dataclass
class HandlerContext:
    """Context object provides knowledge handlers access to other components."""
    
    log: Logger
    identity: NodeIdentity
    config: ConfigProvider | BaseNodeConfig
    cache: Cache
    event_queue: EventQueue
    kobj_queue: KobjQueue
    graph: NetworkGraph
    request_handler: RequestHandler
    resolver: NetworkResolver
    effector: "Effector" = field(init=False)
    
    def set_effector(self, effector: "Effector"):
        """Post initialization injection of effector component."""
        self.effector = effector


from dataclasses import dataclass
from logging import Logger
from rid_lib.ext import Bundle

from ..build.component import depends_on
from ..identity import NodeIdentity
from ..processor.kobj_queue import KobjQueue


@dataclass
class ProfileMonitor:
    """Processes changes to node profile in the config."""
    
    log: Logger
    kobj_queue: KobjQueue
    identity: NodeIdentity
    
    @depends_on("kobj_worker", "port_manager")
    def start(self):
        self.process_profile()
        
    def process_profile(self):
        """Processes identity bundle generated from config."""
        
        self_bundle = Bundle.generate(
            rid=self.identity.rid,
            contents=self.identity.profile.model_dump()
        )
        
        self.kobj_queue.push(bundle=self_bundle)
        
        self.log.debug("Waiting for profile to be processed...")
        # IMPORTANT: this waits for the identity bundle to be processed, later 
        # components (like the handshaker) assume this exists at runtime.
        self.kobj_queue.wait()
        self.log.debug("Done!")
        


from dataclasses import dataclass
from logging import Logger
from rid_lib.ext import Cache
from rid_lib.types import KoiNetNode

from ..build.component import depends_on
from ..network.graph import NetworkGraph
from ..config.base import BaseNodeConfig
from ..identity import NodeIdentity
from ..network.event_queue import EventQueue
from ..protocol.event import Event, EventType


@dataclass
class Handshaker:
    """Handles handshaking with other nodes."""
    
    log: Logger
    cache: Cache
    identity: NodeIdentity
    event_queue: EventQueue
    config: BaseNodeConfig
    graph: NetworkGraph
    
    @depends_on("graph", "profile_monitor", "server", "event_worker")
    def start(self):
        """Attempts handshake with first contact on startup.
        
        Handshake occurs if first contact is set in the config, the first
        contact is not already known to this node, and this node does not
        already have incoming edges with node providers.
        """
        if not self.config.koi_net.first_contact.rid:
            return
        
        if self.cache.read(self.config.koi_net.first_contact.rid):
            return
        
        if self.graph.get_neighbors(
            direction="in", allowed_type=KoiNetNode):
            return
        
        self.handshake_with(self.config.koi_net.first_contact.rid)
        
    def handshake_with(self, target: KoiNetNode):
        """Initiates a handshake with target node.
        
        Pushes successive `FORGET` and `NEW` events to target node to
        reset the target's cache in case it already knew this node. 
        """
        
        self.log.debug(f"Initiating handshake with {target}")
        self.event_queue.push(
            Event.from_rid(
                event_type=EventType.FORGET, 
                rid=self.identity.rid),
            target=target
        )
        self.event_queue.push(
            event=Event.from_bundle(
                event_type=EventType.NEW, 
                bundle=self.cache.read(self.identity.rid)),
            target=target
        )


from dataclasses import dataclass
import socket
from logging import Logger

from ..build.component import depends_on
from ..config.provider import ConfigProvider
from ..config.full_node import FullNodeConfig


@dataclass
class PortManager:
    """Changes port if already in use by another process."""
    
    log: Logger
    config: ConfigProvider | FullNodeConfig
    
    @depends_on("config")
    def start(self):
        self.acquire_port()
        
    def acquire_port(self):
        base_url_is_derived = (self.config.koi_net.node_profile.base_url == self.config.server.url)
        
        changed_port: bool = False
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            address = (self.config.server.host, self.config.server.port)
            while s.connect_ex(address) == 0:
                self.log.debug(f"Port {address[1]} in use")
                self.config.server.port += 1
                address = (address[0], self.config.server.port)
                changed_port = True
        
        self.log.debug(f"Acquired port {address[1]}")
        
        if base_url_is_derived and changed_port:
            self.log.debug("Updating node profile")
            self.config.koi_net.node_profile.base_url = self.config.server.url
        
        self.config.save_to_yaml()


import time
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from fastapi import Request

from ..build.component import depends_on
from ..build.threaded_component import ThreadedComponent
from ..network.response_handler import ResponseHandler
from ..protocol.model_map import API_MODEL_MAP
from ..protocol.api_models import ErrorResponse
from ..protocol.errors import EXCEPTION_TO_ERROR_TYPE, ProtocolError
from ..config.full_node import FullNodeConfig

if TYPE_CHECKING:
    import uvicorn
    from fastapi import FastAPI, APIRouter


@dataclass
class NodeServer(ThreadedComponent):
    """Entry point for full nodes, manages FastAPI server."""
    
    config: FullNodeConfig
    response_handler: ResponseHandler
    
    app: "FastAPI" = field(init=False)
    router: "APIRouter" = field(init=False)
    server: "uvicorn.Server | None" = field(init=False, default=None)
    
    def __post_init__(self):
        self.build_app()
        
    def build_endpoints(self, router: "APIRouter"):
        """Builds endpoints for API router."""
        for path, models in API_MODEL_MAP.items():
            def create_endpoint(path: str):
                async def endpoint(req):
                    return self.response_handler.handle_response(path, req)
                
                # programmatically setting type hint annotations for FastAPI's model validation 
                endpoint.__annotations__ = {
                    "req": models.request_envelope,
                    "return": models.response_envelope
                }
                
                return endpoint
            
            router.add_api_route(
                path=path,
                endpoint=create_endpoint(path),
                methods=["POST"],
                response_model_exclude_none=True
            )
    
    def build_app(self):
        """Builds FastAPI app."""
        from fastapi import FastAPI, APIRouter
        from starlette.middleware.base import BaseHTTPMiddleware

        self.app = FastAPI(
            title="KOI-net Protocol API",
            version="1.1.0"
        )
        
        self.app.add_middleware(BaseHTTPMiddleware, dispatch=self.logging_middleware)
        self.app.add_exception_handler(ProtocolError, handler=self.protocol_error_handler)
        self.router = APIRouter(prefix="/koi-net")
        self.build_endpoints(self.router)
        self.app.include_router(self.router)
    
    async def logging_middleware(self, request: Request, call_next):
        """Binds contextvars per HTTP request, and emits access logs."""
        with self.logging_context.bound_vars(thread="server"):
            self.log.info(f"Request from {request.client.host}:{request.client.port} - {request.method} {request.url.path}")
            response = await call_next(request)
            self.log.info(f"Response code {response.status_code}")
            return response
        
    def protocol_error_handler(self, request, exc: ProtocolError):
        """Catches `ProtocolError` and returns an `ErrorResponse` payload."""
        from fastapi.responses import JSONResponse
        
        self.log.error(exc)
        resp = ErrorResponse(error=EXCEPTION_TO_ERROR_TYPE[type(exc)])
        self.log.info(f"Returning error response: {resp}")
        return JSONResponse(
            status_code=400,
            content=resp.model_dump(mode="json")
        )
    
    def run(self):
        self.server.run()
        
    @depends_on("port_manager")
    def start(self):
        import uvicorn
        self.server = uvicorn.Server(
            config=uvicorn.Config(
            app=self.app,
            host=self.config.server.host,
            port=self.config.server.port,
            log_config=None,
            access_log=False,
            lifespan="off"
        ))
        
        super().start()
        
        deadline = time.monotonic() + 10
        while not self.server.started:
            if time.monotonic() > deadline:
                raise RuntimeError("Server failed to start")
            time.sleep(0.1)
    
    def stop(self):
        if not self.server:
            return
        
        self.server.should_exit = True
        super().stop()


import inspect
from collections import deque
from typing import TYPE_CHECKING, Any

import structlog

from ..exceptions import BuildError
from .component import (
    COMPONENT_TYPE_FIELD,
    DEPENDS_ON_FIELD, 
    START_FUNC_NAME, 
    STOP_FUNC_NAME,
    CompType
)

if TYPE_CHECKING:
    from .assembler import NodeAssembler

log = structlog.stdlib.get_logger()


class BuildArtifact:
    assembler: "NodeAssembler"
    
    comp_dict: dict[str, Any]
    comp_types: dict[str, CompType]
    init_graph: dict[str, set[str]]
    start_graph: dict[str, set[str]]
    stop_graph: dict[str, set[str]]
    
    init_order: list[str]
    start_order: list[str]
    stop_order: list[str]
    
    def __init__(self, assembler: "NodeAssembler"):
        self.assembler = assembler
        
    def collect_components(self):
        """Collects components from class definition."""
        
        self.comp_dict = {}
        # adds components from class and all base classes. skips `type`, and runs in reverse so that sub classes override super class values
        for base in reversed(inspect.getmro(self.assembler)[:-1]):
            for k, v in vars(base).items():
                # excludes built in, private, and `None` attributes
                if k.startswith("_") or v is None:
                    continue
                
                self.comp_dict[k] = v
        log.debug(f"Collected {len(self.comp_dict)} components")
    
    def build_init_graph(self):
        """Builds dependency graph and component type map.
        
        Graph representation is an adjacency list: the key is a component 
        name, and the value is a tuple containing names of the depedencies.
        """
        
        self.comp_types = {}
        self.init_graph = {}
        
        for comp_name, comp in self.comp_dict.items():
            init_dependencies = []
            
            explict_type = getattr(comp, COMPONENT_TYPE_FIELD, None)
            if explict_type:
                self.comp_types[comp_name] = explict_type
                
            elif not callable(comp):
                # non callable components are objects treated "as is"
                self.comp_types[comp_name] = CompType.OBJECT
            else:
                # callable components default to singletons
                self.comp_types[comp_name] = CompType.SINGLETON
            
            if self.comp_types[comp_name] == CompType.SINGLETON:
                sig = inspect.signature(comp)
                init_dependencies = set(sig.parameters)
                
                # difference of sets: dependencies and component names
                # non empty set indicates invalid dependency
                invalid_init_deps = init_dependencies - set(self.comp_dict)
                if invalid_init_deps:
                    log.warning(f"Ignoring undefined init dependencies {invalid_init_deps} on component '{comp_name}'")
                    init_dependencies -= invalid_init_deps
                    
            self.init_graph[comp_name] = init_dependencies
        
        log.debug("Built init dependency graph")
                
    def build_start_graph(self):
        self.start_graph = {}
        start_components = {
            name for name, comp in self.comp_dict.items()
            if getattr(comp, START_FUNC_NAME, None)
        }
        for comp_name, comp in self.comp_dict.items():
            if self.comp_types[comp_name] != CompType.SINGLETON:
                continue
            
            if comp_name not in start_components:
                continue
            
            start_func = getattr(comp, START_FUNC_NAME)
            start_dependencies = getattr(start_func, DEPENDS_ON_FIELD, set())
            invalid_start_deps = start_dependencies - start_components
            if invalid_start_deps:
                log.warning(f"Ignoring undefined start dependencies {invalid_start_deps} on component '{comp_name}'")
                start_dependencies -= invalid_start_deps
                
            self.start_graph[comp_name] = start_dependencies
        
        log.debug("Built start dependency graph")
        
    def build_stop_graph(self):
        self.stop_graph = {}
        
        stop_components = {
            name for name, comp in self.comp_dict.items()
            if getattr(comp, STOP_FUNC_NAME, None)
        }
        
        reverse_start_graph = self.reverse_adj_list(self.start_graph)
        for comp_name, comp in self.comp_dict.items():
            if self.comp_types[comp_name] != CompType.SINGLETON:
                continue
            
            if comp_name not in stop_components:
                continue
            
            # looks for dependencies in this order:
            # @depends_on decorator -> reverse start graph -> empty set
            stop_func = getattr(comp, STOP_FUNC_NAME)
            stop_dependencies = getattr(
                stop_func,
                DEPENDS_ON_FIELD, 
                # default:
                reverse_start_graph.get(
                    comp_name, 
                    # default:
                    set()
                )
            )
            
            invalid_stop_deps = stop_dependencies - stop_components
            if invalid_stop_deps:
                log.warning(f"Ignoring undefined stop dependencies {invalid_stop_deps} on component '{comp_name}'")
                stop_dependencies -= invalid_stop_deps
                
            self.stop_graph[comp_name] = stop_dependencies
        
        log.debug("Built stop dependency graph")
        
    @staticmethod
    def reverse_adj_list(adj: dict[str, set[str]]):
        r_adj: dict[str, set[str]] = {}
        for node in adj:
            r_adj.setdefault(node, set())
            for n in adj[node]:
                r_adj.setdefault(n, set())
                r_adj[n].add(node)
        return r_adj
    
    @staticmethod
    def topo_sort(adj: dict[str, set[str]]):
        """Topological sort of direct graph using Kahn's algorithm."""
        
        # reverse adj list: n -> incoming neighbors
        r_adj = BuildArtifact.reverse_adj_list(adj)
        
        # how many outgoing edges each node has
        out_degree = {
            n: len(neighbors) 
            for n, neighbors in adj.items()
        }
        
        # initializing queue: nodes w/o dependencies
        queue = deque()
        for node in out_degree:
            if out_degree[node] == 0:
                queue.append(node)
        
        ordering = []
        while queue:
            # removes node from graph
            n = queue.popleft()
            ordering.append(n)
            
            # updates out degree for nodes dependent on this node
            for next_n in r_adj[n]:
                out_degree[next_n] -= 1
                # adds nodes now without dependencies to queue
                if out_degree[next_n] == 0:
                    queue.append(next_n)
        
        if len(ordering) != len(adj):
            cycle_nodes = set(adj) - set(ordering)
            raise BuildError(f"Found cycle in dependency graph, the following nodes could not be ordered: {cycle_nodes}")
        
        return ordering
    
    def build_stop_order(self, start_order: list[str]) -> list[str]:
        """Builds component stop order.
        
        Reverse of start order, only including components with a stop method.
        NOTE: Components defining a stop method MUST also define a start method.
        """
        
        stop_order = []
        for comp_name in reversed(start_order):
            comp = self.comp_dict[comp_name]
            if getattr(comp, STOP_FUNC_NAME, None):
                stop_order.append(comp_name)
        
        return stop_order

    @staticmethod
    def visualize(adj: dict[str, list[str]]) -> str:
        """Creates representation of dependency graph in Graphviz DOT language."""
        
        s = "digraph G {\n"
        for node, neighbors in adj.items():
            if node == "graph":
                node = "graph_"
            
            s += f"\t{node};\n"
            for n in neighbors:
                if n == "graph":
                    n = "graph_"
                
                s += f"\t{node} -> {n};\n"
        s += "}"
        return s
    
    def build(self):
        log.debug("Creating build artifact...")
        self.collect_components()
        
        self.build_init_graph()
        log.debug("Starting init graph topo sort...")
        self.init_order = self.topo_sort(self.init_graph)
        log.debug("Init order: " + " -> ".join(self.init_order))
        
        self.build_start_graph()
        log.debug("Starting start graph topo sort...")
        self.start_order = self.topo_sort(self.start_graph)
        log.debug("Start order: " + " -> ".join(self.start_order))
        
        self.build_stop_graph()
        log.debug("Starting stop graph topo sort...")
        self.stop_order = self.topo_sort(self.stop_graph)
        log.debug("Stop order: " + " -> ".join(self.stop_order))
        log.debug("Done")




import time
import threading
from dataclasses import dataclass, field

from ..build.component import depends_on
from ..build.threaded_component import ThreadedComponent
from ..processor.kobj_queue import KobjQueue
from ..network.resolver import NetworkResolver
from ..config.partial_node import PartialNodeConfig


@dataclass
class NodePoller(ThreadedComponent):
    """Entry point for partial nodes, manages polling event loop."""
    
    config: PartialNodeConfig
    kobj_queue: KobjQueue
    resolver: NetworkResolver

    exit_event: threading.Event = field(init=False, default_factory=threading.Event)
    
    def poll(self):
        """Polls neighbor nodes and processes returned events."""
        for node_rid, events in self.resolver.poll_neighbors().items():
            for event in events:
                self.kobj_queue.push(event=event, source=node_rid)

    def run(self):
        """Runs polling event loop."""
        while not self.exit_event.is_set():
            start_time = time.monotonic()
            self.poll()
            elapsed = time.monotonic() - start_time
            wait_time = max(0, self.config.poller.polling_interval - elapsed)
            self.exit_event.wait(wait_time)
    
    @depends_on("graph")
    def start(self):
        self.exit_event.clear()
        super().start()
    
    def stop(self):
        self.exit_event.set()
        super().stop()



import time
from pathlib import Path

from pydantic import BaseModel
from rich.console import Console

from ..config.base import BaseNodeConfig
from ..protocol.node import NodeType
from ..build.container import NodeState
from ..config.provider import ConfigProvider
from .node import NodeInterface


class KoiNetworkConfig(BaseModel):
    first_contact: str | None = None
    nodes: dict[str, str] = {}

class NetworkConfigProvider(ConfigProvider):
    _file_path: str = "koi-network-config.yaml"

class NetworkInterface:
    config: KoiNetworkConfig | NetworkConfigProvider
    
    def __init__(self):
        self.config_schema = KoiNetworkConfig
        self.config = NetworkConfigProvider(
            config_schema=self.config_schema,
            root_dir=Path.cwd()
        )
        
        self.console = Console()
        
        self.nodes: list[NodeInterface] = []
        self.load_nodes()
    
    def load_nodes(self) -> list[NodeInterface]:
        for name, module in self.config.nodes.items():
            node = NodeInterface.from_ref(name, module)
            self.nodes.append(node)
            
            if node.exists():
                # print(f"Loading node '{node.name}'...")
                node.init()
            
    def resolve_node(self, name: str) -> NodeInterface | None:
        for node in self.nodes:
            if node.name == name:
                return node
            
    def add_node(self, node: NodeInterface):
        self.nodes.append(node)
        self.config.nodes[node.name] = node.module
        self.config.save_to_yaml()
        
        fc_node = self.get_first_contact()
        if fc_node:
            with node.mutate_config() as config:
                self.apply_first_contact(
                    source=fc_node.node.config,
                    target=config
                )
    
    def remove_node(self, node: NodeInterface):
        self.nodes.remove(node)
        
        if node.name in self.config.nodes:
            del self.config.nodes[node.name]
            self.config.save_to_yaml()
            
        if self.config.first_contact == node.name:
            self.unset_first_contact(node)
    
    def get_first_contact(self):
        if not self.config.first_contact:
            return
        
        return self.resolve_node(self.config.first_contact)
    
    def apply_first_contact(self, source: BaseNodeConfig, target: BaseNodeConfig):
        target.koi_net.first_contact.rid = source.koi_net.node_rid
        target.koi_net.first_contact.url = source.koi_net.node_profile.base_url
    
    def set_first_contact(self, fc_node: NodeInterface):
        print("Setting first contact...")
        if fc_node.node.config.koi_net.node_profile.node_type == NodeType.PARTIAL:
            print("Partial nodes cannot be first contacts")
            return
        
        prev_fc_node = self.get_first_contact()
        if prev_fc_node:
            if prev_fc_node == fc_node:
                print(f"First contact is already {fc_node.name}")
                return
            self.unset_first_contact(prev_fc_node)
        
        self.config.first_contact = fc_node.name
        self.config.save_to_yaml()
        
        mutated_nodes: int = 0
        for node in self.nodes:
            if node is fc_node:
                continue
            
            with node.mutate_config() as config:
                self.apply_first_contact(
                    source=fc_node.node.config,
                    target=config
                )
                mutated_nodes += 1
        
        print(f"Updated configuration of {mutated_nodes} node(s)")
    
    def unset_first_contact(self, fc_node: NodeInterface):
        print("Unsetting first contact...")
        if not self.config.first_contact:
            return
        
        self.config.first_contact = None
        self.config.save_to_yaml()
        
        mutated_nodes: int = 0
        for node in self.nodes:
            with node.mutate_config() as config:
                if config.koi_net.first_contact.rid == fc_node.node.config.koi_net.node_rid:
                    config.koi_net.first_contact.rid = None
                    config.koi_net.first_contact.url = None
                    mutated_nodes += 1
        
        print(f"Updated configuration of {mutated_nodes} node(s)")

    def sync(self):
        fc_node = self.get_first_contact()
        if fc_node and not fc_node.exists():
            fc_node.create()
            
        for node in self.nodes:
            if not node.exists():
                node.create()
            
            if fc_node and node is not fc_node:
                with node.mutate_config() as config:
                    self.apply_first_contact(
                        source=fc_node.node.config,
                        target=config
                    )
    
    def wipe_cache(self):
        for node in self.nodes:
            if node.exists():
                node.wipe_cache()
                
    def wipe_logs(self):
        for node in self.nodes:
            if node.exists():
                node.wipe_logs()
    
    def wipe_config(self):
        for node in self.nodes:
            if node.exists():
                node.wipe_config()

    def run(self):
        try:
            self.start()
            
            print(f"Completed startup of {len(self.nodes)} node(s)!")
            print("Press Ctrl + C to quit")
            while any(n.state() is NodeState.RUNNING for n in self.nodes):
                time.sleep(0.5)
                
        except KeyboardInterrupt:
            pass
        
        finally:
            self.stop()
    
    def start(self, block: bool = True):
        for name in self.config.nodes:
            node = self.resolve_node(name)
            if node.state() == NodeState.IDLE:
                node.start(block=block)
            
    def stop(self, block: bool = True):
        for name in reversed(self.config.nodes):
            node = self.resolve_node(name)
            if node.state() == NodeState.RUNNING:
                node.stop(block=block)
        



from dataclasses import dataclass
from pathlib import Path

from structlog.contextvars import bound_contextvars


@dataclass
class LoggingContext:
    root_dir: Path
    
    def bound_vars(self, thread: str):
        return bound_contextvars(log_dir=self.root_dir, thread=thread)


from dataclasses import dataclass, field
from logging import Logger

from rid_lib.types import KoiNetNode

from ..behaviors.handshaker import Handshaker
from ..protocol.errors import ErrorType
from ..protocol.event import EventType
from ..processor.kobj_queue import KobjQueue


@dataclass
class ErrorHandler:
    """Handles network and protocol errors that may occur during requests."""
    log: Logger
    kobj_queue: KobjQueue
    handshaker: Handshaker
    
    timeout_counter: dict[KoiNetNode, int] = field(init=False, default_factory=dict)
    
    def reset_timeout_counter(self, node: KoiNetNode):
        """Reset's a node timeout counter to zero."""
        self.timeout_counter[node] = 0
    
    def handle_connection_error(self, node: KoiNetNode):
        """Drops nodes after timing out three times.
        
        TODO: Need a better heuristic for network state. For example, if
        a node lost connection to the internet, it would quickly forget
        all other nodes.
        """
        self.timeout_counter.setdefault(node, 0)
        self.timeout_counter[node] += 1
        
        self.log.debug(f"{node} has timed out {self.timeout_counter[node]} time(s)")
        
        if self.timeout_counter[node] > 3:
            self.log.debug(f"Exceeded time out limit, forgetting node")
            self.kobj_queue.push(rid=node, event_type=EventType.FORGET)
        
    def handle_protocol_error(
        self, 
        error_type: ErrorType, 
        node: KoiNetNode
    ):
        """Handles protocol errors that occur during network requests.
        
        Attempts handshake when this node is unknown to target.
        """
        
        self.log.info(f"Handling protocol error {error_type} for node {node!r}")
        match error_type:
            case ErrorType.UnknownNode:
                self.log.info("Peer doesn't know me, attempting handshake...")
                self.handshaker.handshake_with(node)
                
            case ErrorType.InvalidKey: ...
            case ErrorType.InvalidSignature: ...
            case ErrorType.InvalidTarget: ...



"""Implementation of default knowledge handlers."""

from rid_lib.ext import Bundle
from rid_lib.ext.utils import sha256_hash
from rid_lib.types import KoiNetNode, KoiNetEdge

from ..exceptions import RequestError
from ..protocol.node import NodeType
from ..protocol.event import Event, EventType
from ..protocol.edge import EdgeProfile, EdgeStatus, EdgeType, generate_edge_bundle
from ..protocol.node import NodeProfile
from .handler import KnowledgeHandler, HandlerType, STOP_CHAIN
from .knowledge_object import KnowledgeObject
from .context import HandlerContext


# RID handlers

@KnowledgeHandler.create(HandlerType.RID)
def basic_rid_handler(ctx: HandlerContext, kobj: KnowledgeObject):
    """Default RID handler.
    
    Blocks external events about this node. Sets normalized event type
    for `FORGET` events.
    """
    
    if (kobj.rid == ctx.identity.rid and kobj.source is not None):
        ctx.log.debug("Externally sourced events about this node not allowed.")
        return STOP_CHAIN
    
    if kobj.event_type == EventType.FORGET:
        kobj.normalized_event_type = EventType.FORGET
        return kobj


# Manifest handlers

@KnowledgeHandler.create(HandlerType.Manifest)
def basic_manifest_handler(ctx: HandlerContext, kobj: KnowledgeObject):
    """Normalized event decider based on manifest and cache state.
    
    Stops processing for manifests which have the same hash, or aren't 
    newer than the cached version. Sets the normalized event type to 
    `NEW` or `UPDATE` depending on whether the RID was previously known.
    """
    
    prev_bundle = ctx.cache.read(kobj.rid)

    if prev_bundle:
        if kobj.manifest.sha256_hash == prev_bundle.manifest.sha256_hash:
            ctx.log.debug("Hash of incoming manifest is same as existing knowledge, ignoring")
            return STOP_CHAIN
        if kobj.manifest.timestamp <= prev_bundle.manifest.timestamp:
            ctx.log.debug("Timestamp of incoming manifest is the same or older than existing knowledge, ignoring")
            return STOP_CHAIN
        
        ctx.log.debug("RID previously known to me, labeling as 'UPDATE'")
        kobj.normalized_event_type = EventType.UPDATE

    else:
        ctx.log.debug("RID previously unknown to me, labeling as 'NEW'")
        kobj.normalized_event_type = EventType.NEW
        
    return kobj


# Bundle handlers

@KnowledgeHandler.create(
    handler_type=HandlerType.Bundle,
    rid_types=(KoiNetNode,),
    event_types=(EventType.NEW, EventType.UPDATE))
def secure_profile_handler(ctx: HandlerContext, kobj: KnowledgeObject):
    """Maintains security of cached node profiles.
    
    Blocks bundles with a mismatching public keys in their node profile
    and RID from being written to cache.
    """
    
    node_profile = kobj.bundle.validate_contents(NodeProfile)
    node_rid: KoiNetNode = kobj.rid
    
    if sha256_hash(node_profile.public_key) != node_rid.hash:
        ctx.log.warning(f"Public key hash mismatch for {node_rid!r}!")
        return STOP_CHAIN
    
    if (node_rid != ctx.identity.rid) and (node_profile.base_url == ctx.identity.profile.base_url):
        ctx.log.warning(f"Ignoring node claiming same base URL: {node_profile.base_url}")
        return STOP_CHAIN

@KnowledgeHandler.create(
    handler_type=HandlerType.Bundle, 
    rid_types=(KoiNetEdge,), 
    event_types=(EventType.NEW, EventType.UPDATE))
def edge_negotiation_handler(ctx: HandlerContext, kobj: KnowledgeObject):
    """Handles edge negotiation process.
    
    Automatically approves proposed edges if they request RID types this 
    node can provide (or KOI node, edge RIDs). Validates the edge type 
    is allowed for the node type (partial nodes cannot use webhooks). If 
    edge is invalid, a `FORGET` event is sent to the other node.
    """

    # only handle incoming events (ignore internal edge knowledge objects)
    if kobj.source is None: 
        return
    
    edge_profile = kobj.bundle.validate_contents(EdgeProfile)

    # indicates peer subscribing to this node
    if edge_profile.source == ctx.identity.rid:
        if edge_profile.status != EdgeStatus.PROPOSED:
            return
        
        ctx.log.debug("Handling edge negotiation")
        
        peer_rid = edge_profile.target
        peer_bundle = ctx.cache.read(peer_rid)
        
        if not peer_bundle:
            ctx.log.warning(f"Peer {peer_rid!r} unknown to me")
            return STOP_CHAIN
        
        peer_profile = peer_bundle.validate_contents(NodeProfile)
        
        # explicitly provided event RID types and (self) node + edge objects
        provided_events = (
            *ctx.identity.profile.provides.event,
            KoiNetNode, KoiNetEdge
        )
        
        abort = False
        if (edge_profile.edge_type == EdgeType.WEBHOOK and 
            peer_profile.node_type == NodeType.PARTIAL):
            ctx.log.debug("Partial nodes cannot use webhooks")
            abort = True
        
        if not set(edge_profile.rid_types).issubset(provided_events):
            not_provided = set(edge_profile.rid_types) - set(provided_events)
            ctx.log.debug(f"Requested RID types {not_provided} not provided by this node")
            abort = True
        
        if abort:
            event = Event.from_rid(EventType.FORGET, kobj.rid)
            ctx.event_queue.push(event, peer_rid)
            return STOP_CHAIN
        
        else:
            ctx.log.debug("Approving proposed edge")
            edge_profile.status = EdgeStatus.APPROVED
            updated_bundle = Bundle.generate(kobj.rid, edge_profile.model_dump())
      
            ctx.kobj_queue.push(bundle=updated_bundle, event_type=EventType.UPDATE)
            return
              
    elif edge_profile.target == ctx.identity.rid:
        if edge_profile.status == EdgeStatus.APPROVED:
            ctx.log.debug("Edge approved by other node!")


# Network handlers

@KnowledgeHandler.create(HandlerType.Network, rid_types=(KoiNetNode,))
def node_contact_handler(ctx: HandlerContext, kobj: KnowledgeObject):
    """Makes contact with providers of RID types of interest.
    
    When an incoming node knowledge object is identified as a provider
    of an RID type of interest, this handler will propose a new edge 
    subscribing to future node events, and fetch existing nodes to catch 
    up to the current state.
    """
    # prevents nodes from attempting to form a self loop
    if kobj.rid == ctx.identity.rid:
        return
    
    node_profile = kobj.bundle.validate_contents(NodeProfile)
    
    available_rid_types = list(
        set(ctx.config.koi_net.rid_types_of_interest) & 
        set(node_profile.provides.event)
    )
    
    if not available_rid_types:
        return
    
    edge_rid = ctx.graph.get_edge(
        source=kobj.rid,
        target=ctx.identity.rid,
    )
    
    # already have an edge established
    if edge_rid:
        prev_edge_bundle = ctx.cache.read(edge_rid)
        edge_profile = prev_edge_bundle.validate_contents(EdgeProfile)
        
        if set(edge_profile.rid_types) == set(available_rid_types):
            # no change in rid types
            return
        
        ctx.log.info(f"Proposing updated edge with node provider {available_rid_types}")
        
        edge_profile.rid_types = available_rid_types
        edge_profile.status = EdgeStatus.PROPOSED
        edge_bundle = Bundle.generate(edge_rid, edge_profile.model_dump())
    
    # no existing edge
    else:
        ctx.log.info(f"Proposing new edge with node provider {available_rid_types}")
        edge_bundle = generate_edge_bundle(
            source=kobj.rid,
            target=ctx.identity.rid,
            rid_types=available_rid_types,
            edge_type=(
                EdgeType.WEBHOOK
                if ctx.identity.profile.node_type == NodeType.FULL
                else EdgeType.POLL
            )
        )
    
    # queued for processing
    ctx.kobj_queue.push(bundle=edge_bundle)
    
    ctx.log.info("Catching up on network state")
    try:
        payload = ctx.request_handler.fetch_rids(
            node=kobj.rid, 
            rid_types=available_rid_types
        )
    except RequestError:
        ctx.log.info("Failed to reach node")
        return
        
    for rid in payload.rids:
        if rid == ctx.identity.rid:
            ctx.log.info("Skipping myself")
            continue
        if ctx.cache.exists(rid):
            ctx.log.info(f"Skipping known RID {rid!r}")
            continue
        
        # marked as external since we are handling RIDs from another node
        # will fetch remotely instead of checking local cache
        ctx.kobj_queue.push(rid=rid, source=kobj.rid)
    ctx.log.info("Done")

@KnowledgeHandler.create(HandlerType.Network)
def basic_network_output_filter(ctx: HandlerContext, kobj: KnowledgeObject):
    """Sets network targets of outgoing event for knowledge object.
    
    Allows broadcasting of all RID types this node is an event provider 
    for (set in node profile), and other nodes have subscribed to. All 
    nodes will also broadcast events about their own (internally sourced) 
    KOI node, and KOI edges that they are part of, regardless of their 
    node profile configuration. Finally, nodes will also broadcast about 
    edges to the other node involved (regardless of if they are subscribed).
    """
    
    involves_this_node = False
    # internally source knowledge objects
    if kobj.source is None:
        if type(kobj.rid) is KoiNetNode:
            if (kobj.rid == ctx.identity.rid):
                involves_this_node = True
        
        elif type(kobj.rid) is KoiNetEdge:
            edge_profile = kobj.bundle.validate_contents(EdgeProfile)
            
            if edge_profile.source == ctx.identity.rid:
                ctx.log.debug(f"Adding edge target '{edge_profile.target!r}' to network targets")
                kobj.network_targets.add(edge_profile.target)
                involves_this_node = True
                
            elif edge_profile.target == ctx.identity.rid:
                ctx.log.debug(f"Adding edge source '{edge_profile.source!r}' to network targets")
                kobj.network_targets.add(edge_profile.source)
                involves_this_node = True
    
    if (type(kobj.rid) in ctx.identity.profile.provides.event or involves_this_node):
        subscribers = ctx.graph.get_neighbors(
            direction="out",
            allowed_type=type(kobj.rid)
        )
        
        ctx.log.debug(f"Updating network targets with '{type(kobj.rid)}' subscribers: {subscribers}")
        kobj.network_targets.update(subscribers)
        
    if kobj.source and kobj.source in kobj.network_targets:
        kobj.network_targets.remove(kobj.source)
        ctx.log.debug(f"Removed event source '{kobj.source}' from network targest")
        
    return kobj

@KnowledgeHandler.create(HandlerType.Final, rid_types=[KoiNetNode])
def forget_edge_on_node_deletion(ctx: HandlerContext, kobj: KnowledgeObject):
    """Removes edges to forgotten nodes."""
    
    if kobj.normalized_event_type != EventType.FORGET:
        return
    
    for edge_rid in ctx.graph.get_edges():
        edge_bundle = ctx.cache.read(edge_rid)
        if not edge_bundle:
            continue
        edge_profile = edge_bundle.validate_contents(EdgeProfile)
        
        if kobj.rid in (edge_profile.source, edge_profile.target):
            ctx.log.debug("Identified edge with forgotten node")
            ctx.kobj_queue.push(rid=edge_rid, event_type=EventType.FORGET)


from pathlib import Path

from .build.base import BaseAssembly
from .log_system import LogSystem
from .cache import Cache
from .config.base import BaseNodeConfig
from .config.provider import ConfigProvider
from .config.full_node import FullNodeConfig
from .config.partial_node import PartialNodeConfig
from .processor.context import HandlerContext
from .effector import DerefHandler, Effector
from .behaviors.handshaker import Handshaker
from .behaviors.sync_manager import SyncManager
from .behaviors.port_manager import PortManager
from .identity import NodeIdentity
from .workers import KnowledgeProcessingWorker, EventProcessingWorker
from .network.error_handler import ErrorHandler
from .network.event_queue import EventQueue
from .network.graph import NetworkGraph
from .network.request_handler import RequestHandler
from .network.resolver import NetworkResolver
from .network.response_handler import ResponseHandler
from .network.event_buffer import EventBuffer
from .processor.pipeline import KnowledgePipeline
from .processor.kobj_queue import KobjQueue
from .processor.handler import KnowledgeHandler
from .secure_manager import SecureManager
from .behaviors.profile_monitor import ProfileMonitor
from .entrypoints import NodeServer, NodePoller
from .processor.knowledge_handlers import (
    basic_manifest_handler, 
    basic_network_output_filter, 
    basic_rid_handler, 
    node_contact_handler, 
    edge_negotiation_handler, 
    forget_edge_on_node_deletion, 
    secure_profile_handler
)


class BaseNode(BaseAssembly):
    _log_system: LogSystem = LogSystem
    kobj_queue: KobjQueue = KobjQueue
    event_queue: EventQueue = EventQueue
    poll_event_buf: EventBuffer = EventBuffer
    broadcast_event_buf: EventBuffer = EventBuffer
    config_schema = BaseNodeConfig
    config: ConfigProvider | BaseNodeConfig = ConfigProvider
    knowledge_handlers: list[KnowledgeHandler] = [
        basic_rid_handler,
        basic_manifest_handler,
        secure_profile_handler,
        edge_negotiation_handler,
        node_contact_handler,
        basic_network_output_filter,
        forget_edge_on_node_deletion
    ]
    deref_handlers: list[DerefHandler] = []
    cache: Cache = Cache
    identity: NodeIdentity = NodeIdentity
    graph: NetworkGraph = NetworkGraph
    secure_manager: SecureManager = SecureManager
    handshaker: Handshaker = Handshaker
    error_handler: ErrorHandler = ErrorHandler
    request_handler: RequestHandler = RequestHandler
    sync_manager: SyncManager = SyncManager
    response_handler: ResponseHandler = ResponseHandler
    resolver: NetworkResolver = NetworkResolver
    handler_context: HandlerContext = HandlerContext
    effector: Effector = Effector
    pipeline: KnowledgePipeline = KnowledgePipeline
    kobj_worker: KnowledgeProcessingWorker = KnowledgeProcessingWorker
    event_worker: EventProcessingWorker = EventProcessingWorker
    profile_monitor: ProfileMonitor = ProfileMonitor
    
    def __new__(cls, *args, root_dir: Path = Path.cwd(), **kwargs):
        cls._log_system()
        return super().__new__(cls, *args, root_dir=root_dir, **kwargs)

class FullNode(BaseNode):
    config: FullNodeConfig
    port_manager: PortManager = PortManager
    server: NodeServer = NodeServer

class PartialNode(BaseNode):
    config: PartialNodeConfig
    poller: NodePoller = NodePoller



from pydantic import BaseModel
from .base import BaseNodeConfig, KoiNetConfig as BaseKoiNetConfig
from ..protocol.node import (
    NodeProfile as BaseNodeProfile, 
    NodeType, 
    NodeProvides
)


class NodeProfile(BaseNodeProfile):
    """Node profile config class for partial nodes."""
    base_url: str | None = None
    node_type: NodeType = NodeType.PARTIAL

class KoiNetConfig(BaseKoiNetConfig):
    """KOI-net config class for partial nodes."""
    node_profile: NodeProfile

class PollerConfig(BaseModel):
    """Poller config for partial nodes."""
    polling_interval: int = 5

class PartialNodeConfig(BaseNodeConfig):
    """Node config class for partial nodes."""
    koi_net: KoiNetConfig
    poller: PollerConfig = PollerConfig()


import threading
from dataclasses import dataclass, field
from logging import Logger

from ..logging_context import LoggingContext


@dataclass
class ThreadedComponent:
    """Base class for threaded component. Derived classes MUST set `self.root_dir`."""
    
    log: Logger
    logging_context: LoggingContext
    shutdown_signal: threading.Event
    
    thread: threading.Thread | None = field(init=False, default=None)
    
    def start(self):
        if self.thread and self.thread.is_alive():
            self.log.debug(f"Component {self.__class__.__name__} has already started")
            return
            
        self.thread = threading.Thread(target=self.run_with_log_ctx)
        self.thread.start()
        
    def stop(self):
        if self.thread and self.thread.is_alive():
            self.thread.join()
        else:
            self.log.debug(f"Component {self.__class__.__name__} has already stopped")
    
    def run_with_log_ctx(self):
        with self.logging_context.bound_vars(thread=self.__class__.__name__):
            try:
                self.run()
            except Exception as exc:
                self.log.error(str(exc))
                self.log.error("Raising shutdown signal")
                self.shutdown_signal.set()
    
    def run(self):
        """Processing loop for thread."""
        pass


import time
from typing import Any
import threading
from pathlib import Path
from logging import Logger

from ..lifecycle import NodeLifecycle, NodeState
from ..logging_context import LoggingContext
from .artifact import BuildArtifact


class NodeContainer:
    """Dummy 'shape' for node containers built by assembler."""
    _artifact: BuildArtifact
    _lifecyle: NodeLifecycle

    # components:
    log: Logger
    logging_context: LoggingContext
    root_dir: Path
    shutdown_signal: threading.Event
    
    def __init__(self, artifact, components: dict[str, Any]):
        self._artifact = artifact
        
        # adds all components as attributes of this instance
        for name, comp in components.items():
            setattr(self, name, comp)
            
        
        self._lifecycle = NodeLifecycle(
            shutdown_signal=self.shutdown_signal,
            log=self.log,
            logging_context=self.logging_context,
            artifact=self._artifact,
            container=self
        )
    
    def run(self):
        try:
            self.start()
            print("Press Ctrl + C to quit")
            while (self.get_state() is NodeState.RUNNING):
                time.sleep(0.5)
        
        except KeyboardInterrupt:
            print("Quitting...")
        
        finally:
            self.stop()
    
    def start(self, block: bool = True):
        self._lifecycle.start(block=block)
        
    def stop(self, block: bool = True):
        self._lifecycle.stop(block=block)
        
    def get_state(self) -> NodeState:
        return self._lifecycle.state


from dataclasses import dataclass
from logging import Logger
from typing import Callable
from enum import StrEnum

from rid_lib.ext import Cache, Bundle
from rid_lib.core import RID, RIDType
from rid_lib.types import KoiNetNode

from .processor.context import HandlerContext
from .network.resolver import NetworkResolver
from .processor.kobj_queue import KobjQueue


@dataclass
class DerefHandler:
    func: Callable[[HandlerContext, RID], Bundle | None]
    rid_types: tuple[RIDType]
    
    def __call__(self, ctx: HandlerContext, rid: RID) -> Bundle | None:
        return self.func(ctx, rid)
    
    @classmethod
    def create(cls, rid_types: tuple[RIDType]):
        def decorator(func: Callable) -> DerefHandler:
            handler = cls(func, rid_types)
            return handler
        return decorator

class BundleSource(StrEnum):
    CACHE = "CACHE"
    ACTION = "ACTION"

@dataclass
class Effector:
    """Subsystem for dereferencing RIDs."""
    
    log: Logger
    cache: Cache
    resolver: NetworkResolver
    kobj_queue: KobjQueue
    handler_context: HandlerContext
    deref_handlers: list[DerefHandler]
    
    def __post_init__(self):
        self.handler_context.set_effector(self)
    
    def _try_cache(self, rid: RID) -> tuple[Bundle, BundleSource] | None:
        bundle = self.cache.read(rid)
        
        if bundle:
            self.log.debug("Cache hit")
            return bundle, BundleSource.CACHE
        else:
            self.log.debug("Cache miss")
            return None
            
    def _try_action(self, rid: RID) -> tuple[Bundle, BundleSource] | None:
        action = None
        for handler in self.deref_handlers:
            if type(rid) not in handler.rid_types:
                continue
            action = handler
            break
        
        if not action:
            self.log.debug("No action found")
            return None
        
        bundle = action(ctx=self.handler_context, rid=rid)
        
        if bundle:
            self.log.debug("Action hit")
            return bundle, BundleSource.ACTION
        else:
            self.log.debug("Action miss")
            return None
        
    def _try_network(self, rid: RID) -> tuple[Bundle, KoiNetNode] | None:
        bundle, source = self.resolver.fetch_remote_bundle(rid)
        
        if bundle:
            self.log.debug("Network hit")
            return bundle, source
        else:
            self.log.debug("Network miss")
            return None
        
    def deref(
        self, 
        rid: RID,
        refresh_cache: bool = False,
        use_network: bool = False,
        handle_result: bool = True,
        write_through: bool = False
    ) -> Bundle | None:
        """Dereferences an RID.
        
        Attempts to dereference an RID by (in order) reading the cache, 
        calling a bound action, or fetching from other nodes in the 
        newtork.
        
        Args:
            rid: RID to dereference
            refresh_cache: skips cache read when `True` 
            use_network: enables fetching from other nodes when `True`
            handle_result: sends resulting bundle to kobj queue when `True`
            write_through: waits for kobj queue to empty when `True`
        """
        
        self.log.debug(f"Dereferencing {rid!r}")
        
        bundle, source = (
            # if `refresh_cache`, skip try cache
            not refresh_cache and self._try_cache(rid) or 
            self._try_action(rid) or
            use_network and self._try_network(rid) or
            # if not found, bundle and source set to None
            (None, None) 
        )
        
        if (
            handle_result 
            and bundle is not None 
            and source != BundleSource.CACHE
        ):            
            self.kobj_queue.push(
                bundle=bundle, 
                source=source if type(source) is KoiNetNode else None
            )
            
            if write_through:
                self.kobj_queue.wait()
                
        return bundle


import importlib
import inspect
import pkgutil
from importlib.metadata import entry_points
from types import ModuleType

from ..core import BaseNode


ENTRY_POINT_GROUP = "koi_net.node"
MODULE_PREFIX = "koi_net_"
MODULE_POSTFIX = "_node"
MODULE_CORE = ".core"


class ModuleInterface:
    def __init__(self):
        self.module_names: set[str] = set()
        # alias -> module name
        self.alias_module_map: dict[str, str] = {}
        self.module_map: dict[str, ModuleType] = {}
        
        self.load_module_names()
        
    def resolve_ref(self, module_ref) -> str:
        if module_ref in self.module_names:
            module_name = module_ref
        elif module_ref in self.alias_module_map:
            module_name = self.alias_module_map[module_ref]
        else:
            raise ModuleNotFoundError(f"Couldn't resolve module reference '{module_ref}'")
        
        return module_name
    
    def load_class(self, module_name: str, reload_module: bool = False):
        if module_name not in self.module_map:
            print(f"Loading module {module_name}...")
            module = importlib.import_module(module_name + MODULE_CORE)
            self.module_map[module_name] = module
        elif reload_module:
            print(f"Reloading module {module_name}...")
            module = importlib.reload(self.module_map[module_name])
            self.module_map[module_name] = module
        else:
            module = self.module_map[module_name]
        
        for _, obj in inspect.getmembers(module):
            if (
                getattr(obj, "__module__", None) == module.__name__ and 
                inspect.isclass(obj) and
                issubclass(obj, BaseNode)
            ):
                return obj
        
    def load_module_names(self):
        for ep in entry_points(group=ENTRY_POINT_GROUP):
            self.module_names.add(ep.module)
            self.alias_module_map[ep.name] = ep.module

        for module in pkgutil.iter_modules():
            if (module.name.startswith(MODULE_PREFIX) and module.name.endswith(MODULE_POSTFIX)):
                self.module_names.add(module.name)
                module_alias = module.name[len(MODULE_PREFIX):-len(MODULE_POSTFIX)]
                self.alias_module_map.setdefault(module_alias, module.name)

module_interface = ModuleInterface()


import time
import os
import shutil
from pathlib import Path
from typing import Any

from jsonpointer import JsonPointer, JsonPointerException
from pydantic import ValidationError
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich import box

from rid_lib.types import KoiNetEdge, KoiNetNode

from koi_net.exceptions import MissingEnvVarsError
from ..protocol.edge import EdgeProfile

from ..core import BaseNode
from ..build.container import NodeState, NodeContainer
from ..log_system import LogSystem
from .module import module_interface


class NodeInterface:
    node: BaseNode | NodeContainer
    
    def __init__(
        self,
        name: str,
        node_class: type[BaseNode]
    ):
        self.name = name
        self.node_class = node_class
        # top level module
        self.module = node_class.__module__.split(".")[0]
        self._node = None
        self.initialized = False
        
        self.console = Console()
        
    @classmethod
    def from_ref(cls, name: str, module_ref: str):
        module = module_interface.resolve_ref(module_ref)
        node_class = module_interface.load_class(module)
        return cls(name=name, node_class=node_class)
    
    @property
    def node(self):
        if not self.initialized:
            raise Exception("node not initialized")

        return self._node
    
    def set_node_class(self, node_class):
        self.node_class = node_class
        self.initialized = False
        self.init()
    
    def create(self):
        os.mkdir(self.name)
        self.init()
        if self.initialized:
            self.console.print(f"Created node '{self.node.identity.rid}'")
        else:
            # self.console.print("Failed to initialize, run `node init <name>` to try again")
            self.delete()
    
    def exists(self):
        return os.path.isdir(self.name)
    
    def delete(self):
        LogSystem.delete_file_handler(self.name)
        shutil.rmtree(self.name)
        
    def init(self):
        if not self.initialized:
            try:
                self._node = self.node_class(root_dir=Path(self.name))
                self.initialized = True
            except MissingEnvVarsError as err:
                text = "\n".join([
                    f"[bold red]{v}[/bold red]" 
                    for v in err.vars
                ])
                
                self.console.print(
                    Panel.fit(
                        renderable=text, 
                        title="Cannot initialize node, missing the following enironment variables:",
                        border_style="red"))
            except Exception as err:
                print(err)
    
    def state(self):
        return self.node.get_state()
    
    def info(self):
        table = Table("rid types", "", "node", box=box.SIMPLE)
        for edge_rid in self.node.cache.list_rids(rid_types=[KoiNetEdge]):
            bundle = self.node.cache.read(edge_rid)
            edge_profile = bundle.validate_contents(EdgeProfile)
            
            if edge_profile.source == self.node.identity.rid:
                direction = "->"
                other = edge_profile.target
                
            elif edge_profile.target == self.node.identity.rid:
                direction = "<-"
                other = edge_profile.source
            
            else:
                continue
            
            table.add_row(str(edge_profile.rid_types[0]), direction, str(other))
            for rid_type in edge_profile.rid_types[1:]:
                table.add_row(str(rid_type))
            table.add_row()
        
        self.console.print(table)
            
        
    def run(self):
        try:
            self.start()
            print("Press Ctrl + C to quit")
            while(self.state() is NodeState.RUNNING):
                time.sleep(0.5)
        
        except KeyboardInterrupt:
            pass
        
        finally:
            self.stop()
        
    def start(self, block: bool = True):
        if self.state() == NodeState.IDLE:
            print(f"Starting {self.name}...", end=" ", flush=True)
            self.node.start(block=block)
            print("Done")
        else:
            print("Node already started")
        
    def stop(self, block: bool = True):
        if self.state() == NodeState.RUNNING:
            print(f"Stopping {self.name}...", end=" ", flush=True)
            self.node.stop(block=block)
            print("Done")
        else:
            print("Node already stopped")
    
    def wipe_config(self):
        self.node.config.wipe()
    
    def wipe_cache(self):
        self.node.cache.drop()
        
    def wipe_logs(self):
        LogSystem.delete_file_handler(self.name, wipe_logs=True)
        
    def mutate_config(self):
        return self.node.config.mutate()
        
    def get_config(self, jp: str) -> Any:
        config_json = self.node.config.model_dump(mode="json")
        try:
            val = JsonPointer(jp).get(config_json)
            return val
        except JsonPointerException:
            self.console.print(f"[bold red]Location '{jp}' does not exist[/bold red]")
            raise KeyError("Invalid JSON pointer")
    
    def set_config(self, jp: str, val: Any):
        data = self.node.config.model_dump()
        pointer = JsonPointer(jp)
        prev_val = pointer.get(data)
        pointer.set(data, val)
        
        try:
            config = self.node.config_schema.model_validate(data)
        except ValidationError as err:
            self.console.print(f"[bold red]Invalid value '{val}': {err.errors()[0]['msg']}[/bold red]")
            return
            
        self.node.config._set_delegate(config)
        self.node.config.save_to_yaml()
        
        val_repr = lambda v: f"'{v}'" if v is not None else "<null>"
        
        self.console.print(f"Set config value [cyan]{val_repr(prev_val)}[/cyan] -> [green]{val_repr(val)}[/green]")
        
    def unset_config(self, jp: str):
        self.set_config(jp, None)
    


import inspect
import os
from pathlib import Path
from contextlib import contextmanager

from pydantic import ValidationError

from ..exceptions import MissingEnvVarsError
from .env_config import EnvConfig
from .base import BaseNodeConfig


DELEGATE = "_delegate"

class ConfigProvider:
    """Loads node config from a YAML file, and proxies access to it."""
    
    _file_path: str = "config.yaml"
    _file_content: str
    
    _schema: type[BaseNodeConfig]
    _root_dir: Path
    
    _delegate: BaseNodeConfig
    
    def __init__(self, config_schema, root_dir):
        object.__setattr__(self, "_schema", config_schema)
        object.__setattr__(self, "_root_dir", root_dir)
        
        # this is a special case to allow config state dependent components
        # to initialize without a "lazy initialization" approach, in general
        # components SHOULD NOT execute code in their init phase
        
        self._validate_env_vars()
        self._load_from_yaml()
        
    def _set_delegate(self, delegate: BaseNodeConfig):
        object.__setattr__(self, DELEGATE, delegate)
        
    def _get_delegate(self) -> BaseNodeConfig:
        return object.__getattribute__(self, DELEGATE)
    
    def __getattr__(self, name):
        return getattr(self._get_delegate(), name)
    
    def __setattr__(self, name, value):
        """Overrides set attribute for ALL members of this class.
        
        Any non proxying set attribute call needs to be done using
        `object.__attribute__(self, name)`.
        """
        
        delegate = self._get_delegate()
        setattr(delegate, name, value)

    def _validate_env_vars(self):
        """Validates environment variables and raises formatted exception.
        
        Useful for interfacing with CLI, catch exception and print the 
        missing vars to the screen.
        """
        for field in self._schema.model_fields.values():
            field_type = field.annotation
            if inspect.isclass(field_type) and issubclass(field_type, EnvConfig):
                try:
                    field_type()
                except ValidationError as exc:
                    missing_vars = [
                        err["loc"][0].upper()
                        for err in exc.errors()
                        if err["type"] == "missing"
                    ]
                    raise MissingEnvVarsError(
                        f"Missing required vars: {','.join(v for v in missing_vars)}",
                        vars=missing_vars)
    
    def _load_from_yaml(self):
        """Loads config from YAML file, or generates it if missing."""
        from ruamel.yaml import YAML
        yaml = YAML()
        
        try:
            # loads from yaml
            with open(self._root_dir / self._file_path, "r") as f:
                object.__setattr__(self, "_file_content", f.read())
            config_data = yaml.load(self._file_content)
            config = self._schema.model_validate(config_data)
        
        except FileNotFoundError:
            # loads defaults
            config = self._schema()
        
        # loads to delegate
        self._set_delegate(config)
        
    def save_to_yaml(self):
        """Saves config to YAML file."""
        from ruamel.yaml import YAML
        yaml = YAML()
        
        with open(self._root_dir / self._file_path, "w") as f:
            try:
                config = self._get_delegate()
                config_data = config.model_dump(
                    mode="json",
                    exclude={"env": True})
                yaml.dump(config_data, f)
                
            except Exception:
                # rewrites original content if YAML dump fails
                if self._file_content:
                    f.seek(0)
                    f.truncate()
                    f.write(self._file_content)
                raise
            
    def wipe(self):
        try:
            os.remove(self._root_dir / self._file_path)
        except FileNotFoundError:
            pass
        
        try:
            os.remove(self._root_dir / self.private_key_pem_path)
        except FileNotFoundError:
            pass
    
    @contextmanager
    def mutate(self):
        yield self._get_delegate()
        self.save_to_yaml()
        
    def start(self):
        self.save_to_yaml()


from dataclasses import dataclass
from logging import Logger
from rid_lib.types import KoiNetEdge, KoiNetNode
from rid_lib.ext import Cache

from ..exceptions import RequestError
from ..protocol.event import EventType
from ..network.request_handler import RequestHandler
from ..network.event_queue import EventQueue
from ..network.graph import NetworkGraph
from .handler import (
    KnowledgeHandler,
    HandlerType, 
    STOP_CHAIN,
    StopChain
)
from .knowledge_object import KnowledgeObject
from .context import HandlerContext


@dataclass
class KnowledgePipeline:
    log: Logger
    handler_context: HandlerContext
    cache: Cache
    request_handler: RequestHandler
    event_queue: EventQueue
    graph: NetworkGraph
    knowledge_handlers: list[KnowledgeHandler]
    
    def call_handler_chain(
        self, 
        handler_type: HandlerType,
        kobj: KnowledgeObject
    ) -> KnowledgeObject | StopChain:
        """Calls handlers of provided type, chaining their inputs and outputs together.
        
        The knowledge object provided when this function is called will be passed to the first handler. A handler may return one of three types: 
        - `KnowledgeObject` - to modify the knowledge object for the next handler in the chain
        - `None` - to keep the same knowledge object for the next handler in the chain
        - `STOP_CHAIN` - to stop the handler chain and immediately exit the processing pipeline
        
        Handlers will only be called in the chain if their handler and RID type match that of the inputted knowledge object. 
        """
        
        for handler in self.knowledge_handlers:
            if handler_type != handler.handler_type: 
                continue
            
            if handler.rid_types and type(kobj.rid) not in handler.rid_types:
                continue
            
            if handler.event_types and kobj.event_type not in handler.event_types:
                continue
            
            self.log.debug(f"Calling {handler_type} handler '{handler.func.__name__}'")
            
            resp = handler(
                ctx=self.handler_context,
                kobj=kobj.model_copy()
            )
            
            # stops handler chain execution
            if resp is STOP_CHAIN:
                self.log.debug(f"Handler chain stopped by {handler.func.__name__}")
                return STOP_CHAIN
            
            # kobj unmodified
            elif resp is None:
                continue
            
            # kobj modified by handler
            elif isinstance(resp, KnowledgeObject):
                kobj = resp
                self.log.debug(f"Knowledge object modified by {handler.func.__name__}")
            
            else:
                raise ValueError(f"Handler {handler.func.__name__} returned invalid response '{resp}'")
            
        return kobj
    
    def process(self, kobj: KnowledgeObject):
        """Sends knowledge object through knowledge processing pipeline.
        
        Handler chains are called in between major events in the 
        pipeline, indicated by their handler type. Each handler type is 
        guaranteed to have access to certain knowledge, and may affect a 
        subsequent action in the pipeline. The five handler types are as 
        follows:
        - RID - provided RID; if event type is `FORGET`, this handler 
        decides whether to delete the knowledge from the cache by 
        setting the normalized event type to `FORGET`, otherwise this 
        handler decides whether to validate the manifest (and fetch it 
        if not provided). After processing, if event type is `FORGET`, 
        the manifest and contents will be retrieved from the local cache, 
        and indicate the last state of the knowledge before it was 
        deleted.
        - Manifest - provided RID, manifest; decides whether to validate 
        the bundle (and fetch it if not provided).
        - Bundle - provided RID, manifest, contents (bundle); decides 
        whether to write knowledge to the cache by setting the 
        normalized event type to `NEW` or `UPDATE`.
        - Network - provided RID, manifest, contents (bundle); decides 
        which nodes (if any) to broadcast an event about this knowledge 
        to.
        - Final - provided RID, manifests, contents (bundle); final 
        action taken after network broadcast.
        
        The pipeline may be stopped by any point by a single handler 
        returning the `STOP_CHAIN` sentinel. In that case, the process 
        will exit immediately. Further handlers of that type and later 
        handler chains will not be called.
        """
        
        self.log.debug(f"Handling {kobj!r}")
        kobj = self.call_handler_chain(HandlerType.RID, kobj)
        if kobj is STOP_CHAIN: return
        
        if kobj.event_type == EventType.FORGET:
            bundle = self.cache.read(kobj.rid)
            if not bundle:
                self.log.debug("Local bundle not found")
                return
            
            # the bundle (to be deleted) attached to kobj for downstream analysis
            self.log.debug("Adding local bundle (to be deleted) to knowledge object")
            kobj.manifest = bundle.manifest
            kobj.contents = bundle.contents
            
        else:
            # attempt to retrieve manifest
            if not kobj.manifest:
                self.log.debug("Manifest not found")
                if not kobj.source:
                    return
            
                self.log.debug("Attempting to fetch remote manifest from source")
                try:
                    payload = self.request_handler.fetch_manifests(
                        node=kobj.source,
                        rids=[kobj.rid])
                except RequestError:
                    self.log.debug("Failed to find manifest")
                    return
                
                kobj.manifest = payload.manifests[0]
                
            kobj = self.call_handler_chain(HandlerType.Manifest, kobj)
            if kobj is STOP_CHAIN: return
            
            # attempt to retrieve bundle
            if not kobj.contents:
                self.log.debug("Bundle not found")
                if kobj.source is None:
                    return
                
                self.log.debug("Attempting to fetch remote bundle from source")
                try:
                    payload = self.request_handler.fetch_bundles(
                        node=kobj.source,
                        rids=[kobj.rid]
                    )
                except RequestError:
                    self.log.debug("Failed to find bundle")
                    return
                
                bundle = payload.bundles[0]
                
                if kobj.manifest != bundle.manifest:
                    self.log.warning("Retrieved bundle contains a different manifest")
                
                kobj.manifest = bundle.manifest
                kobj.contents = bundle.contents
                
        kobj = self.call_handler_chain(HandlerType.Bundle, kobj)
        if kobj is STOP_CHAIN: return
            
        if kobj.normalized_event_type in (EventType.UPDATE, EventType.NEW):
            self.log.info(f"Writing to cache: {kobj!r}")
            self.cache.write(kobj.bundle)
            
        elif kobj.normalized_event_type == EventType.FORGET:
            self.log.info(f"Deleting from cache: {kobj!r}")
            self.cache.delete(kobj.rid)
            
        else:
            self.log.debug("Normalized event type was not set, no cache or network operations will occur")
            return
        
        if type(kobj.rid) in (KoiNetNode, KoiNetEdge):
            self.log.debug("Change to node or edge, regenerating network graph")
            self.graph.generate()
        
        kobj = self.call_handler_chain(HandlerType.Network, kobj)
        if kobj is STOP_CHAIN: return
        
        if kobj.network_targets:
            self.log.debug(f"Broadcasting event to {len(kobj.network_targets)} network target(s)")
        else:
            self.log.debug("No network targets set")
        
        for node in kobj.network_targets:
            self.event_queue.push(kobj.normalized_event, node)
        
        kobj = self.call_handler_chain(HandlerType.Final, kobj)



import threading
from dataclasses import dataclass, field
from logging import Logger
from enum import StrEnum
from typing import Any

from koi_net.build.artifact import BuildArtifact
from koi_net.logging_context import LoggingContext

from .build.component import START_FUNC_NAME, STOP_FUNC_NAME


class NodeState(StrEnum):
    IDLE = "IDLE"
    STARTING = "STARTING"
    RUNNING = "RUNNING"
    STOPPING = "STOPPING"

@dataclass
class NodeLifecycle:
    
    # injected components:
    shutdown_signal: threading.Event
    log: Logger
    logging_context: LoggingContext
    artifact: BuildArtifact
    container: Any
    
    # internal vars:
    err: Exception = field(init=False, default=None)
    state: NodeState = field(init=False, default=NodeState.IDLE)
    thread: threading.Thread | None = field(init=False, default=None)
    startup_signal: threading.Event = field(init=False, default_factory=threading.Event)
    
    def start(self, block: bool = True):
        if self.state != NodeState.IDLE:
            self.log.warning("Node can't be started from non-idle state")
            return
        
        self.startup_signal.clear()
        self.thread = threading.Thread(target=self.run)
        self.thread.start()
        
        if block:
            self.startup_signal.wait()
        
    def stop(self, block: bool = True):
        if self.state != NodeState.RUNNING:
            self.log.warning("Node can't be stopped from non-running state")
            return
        
        self.shutdown_signal.set()
        
        if block and self.thread and self.thread.is_alive():
            self.thread.join()

    def run(self):
        with self.logging_context.bound_vars(thread=self.__class__.__name__):
            try:
                self.startup()
                self.startup_signal.set()
                self.shutdown_signal.wait()
                
            finally:
                self.shutdown()
                if self.err:
                    raise self.err
        
    def startup(self):
        self.state = NodeState.STARTING
        self.log.info("Starting node...")
        for comp_name in self.artifact.start_order:
            comp = getattr(self.container, comp_name)
            start_func = getattr(comp, START_FUNC_NAME)
            self.log.info(f"Starting {comp_name}...")
            
            try:
                start_func()
            except Exception as err:
                self.shutdown_signal.set()
                self.log.error(str(err))
                self.err = err
            
            if self.shutdown_signal.is_set():
                self.log.error(f"Startup failed, aborting")
                return
        
        self.state = NodeState.RUNNING
        self.log.info("Startup complete!")
            
    def shutdown(self):
        self.state = NodeState.STOPPING
        self.log.info("Stopping node...")
        for comp_name in self.artifact.stop_order:
            comp = getattr(self.container, comp_name)
            stop_func = getattr(comp, STOP_FUNC_NAME)
            self.log.info(f"Stopping {comp_name}...")
            stop_func()
        
        self.shutdown_signal.clear()
        self.state = NodeState.IDLE
        self.log.info("Shutdown complete!")



import cmd
import inspect
import shlex
from functools import wraps

from rich.console import Console
from rich.table import Table
from rich import box

from ..build.container import NodeState
from ..log_system import LogSystem
from .module import module_interface
from .network import NetworkInterface
from .node import NodeInterface

class KoiShell(cmd.Cmd):
    # intro = "Welcome to the KOI shell, type help for a list of commands.\n"
    prompt = ">=> "
    file = None
    
    def __init__(self):
        super().__init__()
        LogSystem(use_console_handler=False)
        self.console = Console()
        
        self.console.print("KOI shell: [cyan]type `help` for a list of commands[/cyan]")
        
        self.network = NetworkInterface()
        self.console.print(f"Found [green]{len(module_interface.module_names)} module(s)[/green] and [green]{len(self.network.nodes)} node(s)[/green]")
        print()
        
    @staticmethod
    def load_node(func):
        @wraps(func)
        def wrapper(self: "KoiShell", name: str, *args):
            node = self.network.resolve_node(name)
            if node:
                return func(self, node, *args)
            else:
                print(f"Could not find node '{name}'")
        return wrapper
    
    @staticmethod
    def parse_args(func):
        @wraps(func)
        def wrapper(self: "KoiShell", arg: str, *args):
            parsed_args = shlex.split(arg)
            return func(self, *parsed_args)
        return wrapper
    
    @staticmethod
    def validate_args(func):
        @wraps(func)
        def wrapper(self: "KoiShell", *args):
            sig = inspect.signature(func)
            try:
                sig.bind(self, *args)
            except TypeError as err:
                self.console.print(f"[bold red]{str(err).capitalize()}[/bold red]")
                return
            
            return func(self, *args)
        return wrapper
        
    def do_quit(self, arg: str):
        if all(not node.initialized or node.state() == NodeState.IDLE for node in self.network.nodes):
            return True
        print("Nodes are still running! Run `network stop` first, or `QUIT` to exit anyway.")
        
    def do_QUIT(self, arg: str):
        self.network.stop()
        return True
    
    def do_help(self, subcmd: str):
        table = Table("command", "arguments", "info", box=box.SIMPLE)
        match subcmd:
            case "node":
                cmd_help = [
                    ("add", r"<module ref> \[name]", "adds a new node of type <module ref> to the network, if unset, name defaults to <module ref>"),
                    ("rm", "<name>", "removes a node from the network"),
                    ("init", "<name>", "initializes a node"),
                    ("wipe-config", "", "wipes a node's configuration, including private key"),
                    ("wipe-cache", "<name>", "wipes a node's RID cache"),
                    ("wipe-logs", "<name>", "wipes a node's logs"),
                    
                    ("config-get", "<name> <loc>", "prints the config value at the specified JSON pointer location"),
                    ("config-set", "<name> <loc> <val>", "sets the config at JSON pointer location to value"),
                    ("config-unset", "<name> <loc>", ""),
                    ("info", "<name>", "shows info about nodes edges"),
                    
                    ("run", "<name>", "runs a node in the foreground"),
                    ("start", "<name>", "starts a node in the background"),
                    ("stop", "<name>", "stops a running background node"),
                    
                    ("list", "", "lists all nodes in the network")
                ]
            case "network":
                cmd_help = [
                    ("sync", "", "synchronizes the local environment with the network configuration"),
                    ("wipe-config", "", "wipes configuration, including private key, of all network nodes"),
                    ("wipe-cache", "", "wipes RID cache of all network nodes"),
                    ("wipe-logs", "", "wipes logs of all network nodes"),
                    ("status", "", "lists the current state of all network nodes"),
                    
                    ("set-first-contact", "<name>", "Sets first contact of all nodes in the network"),
                    ("unset-first-contact", "", "Unsets first contact from all nodes in the network"),
                    
                    ("run", "", "runs all network nodes in the foreground"),
                    ("start", "", "starts all network nodes in the backround"),
                    ("stop", "", "stops all running background nodes"),
                    
                ]
            case "module":
                cmd_help = [
                    ("list", "", "lists all detected node modules"),
                ]
            case _:
                cmd_help = [
                    ("node", "<sub cmd>", "group of node commands"),
                    ("network", "<sub cmd>", "group of network commands"),
                    ("module", "<sub cmd>", "group of module commands"),
                    ("help", r"\[cmd]", "list available commands or subcommands"),
                    ("quit", "", "exits the shell"),
                    ("QUIT", "", "exits the shell, even if nodes are running"),
                ]
        
        for cmd_info in cmd_help:
            table.add_row(*cmd_info)
        
        self.console.print(table)
    
    @parse_args
    def do_node(self, sub_cmd: str, *args):
        match sub_cmd:
            case "add":
                self.node_add(*args)
            case "rm":
                self.node_rm(*args)
            case "init":
                self.node_init(*args)
            case "list":
                self.node_list(*args)
            case "config-get":
                self.node_config_get(*args)
            case "config-set":
                self.node_config_set(*args)
            case "config-unset":
                self.node_config_unset(*args)
            case "info":
                self.node_info(*args)
            case "wipe-config":
                self.node_wipe_config(*args)
            case "wipe-cache":
                self.node_wipe_cache(*args)
            case "wipe-logs":
                self.node_wipe_logs(*args)
            case "start":
                self.node_start(*args)
            case "stop":
                self.node_stop(*args)
            case "run":
                self.node_run(*args)
            case _:
                print(f"Unknown subcommand '{sub_cmd}'")
    
    @parse_args
    def do_network(self, sub_cmd: str, *args):
        match sub_cmd:
            case "sync":
                self.network_sync(*args)
            case "wipe-config":
                self.network_wipe_config(*args)
            case "wipe-cache":
                self.network_wipe_cache(*args)
            case "wipe-logs":
                self.network_wipe_logs(*args)
            case "status":
                self.network_status(*args)
            case "set-first-contact":
                self.network_set_first_contact(*args)
            case "unset-first-contact":
                self.network_unset_first_contact(*args)
            case "start":
                self.network_start(*args)
            case "stop":
                self.network_stop(*args)
            case "run":
                self.network_run(*args)
            case "set-first-contact":
                self.network_set_first_contact(*args)
            case _:
                print(f"Unknown subcommand '{sub_cmd}'")
                
    @parse_args
    def do_module(self, sub_cmd: str, *args):
        match sub_cmd:
            case "list":
                self.module_list(*args)
            case "reload":
                self.module_reload(*args)
            case _:
                print(f"Unknown subcommand '{sub_cmd}'")
            
    @validate_args
    def node_add(self, module_ref: str, name: str | None = None):
        name = name or module_ref
        if self.network.resolve_node(name):
            print(f"Node with name '{name}' already exists")
        
        try:
            node = NodeInterface.from_ref(name, module_ref)
        except ModuleNotFoundError as err:
            print(err)
            return
        
        if not node.exists():
            node.create()
        
        if node.initialized:
            self.network.add_node(node)
    
    @validate_args
    @load_node
    def node_rm(self, node: NodeInterface):
        if node.exists():
            if node.state() == NodeState.IDLE:
                node.delete()
                print(f"Removed node '{node.name}'")
            else:
                print(f"Node is running, run `node stop {node.name}` first")
                return
        
        self.network.remove_node(node)
        
    @validate_args
    @load_node
    def node_init(self, node: NodeInterface):
        node.init()
    
    @validate_args
    def node_list(self):
        table = Table("name", "module", "rid", box=box.SIMPLE)
        for node in self.network.nodes:
            if not node.exists():
                continue
            
            node_rid = node.node.config.koi_net.node_rid
            table.add_row(node.name, node.module, str(node_rid))
        self.console.print(table)
    
    @validate_args
    @load_node
    def node_config_get(self, node: NodeInterface, loc: str):
        try:
            val = node.get_config(loc)
            self.console.print(val)
        except KeyError:
            pass
        
    @validate_args
    @load_node
    def node_config_set(self, node: NodeInterface, loc: str, val: str):
        node.set_config(loc, val)
    
    @validate_args
    @load_node
    def node_config_unset(self, node: NodeInterface, loc: str):
        node.unset_config(loc)
        
    @load_node
    def node_info(self, node: NodeInterface):
        node.info()
    
    @validate_args
    @load_node
    def node_run(self, node: NodeInterface):
        node.run()
    
    @validate_args
    @load_node
    def node_start(self, node: NodeInterface):
        node.start()
    
    @validate_args
    @load_node
    def node_stop(self, node: NodeInterface):
        node.stop()
    
    @validate_args
    @load_node
    def node_wipe_config(self, node: NodeInterface):
        node.wipe_config()
        print(f"Wiped config of '{node.name}'")
    
    @validate_args
    @load_node
    def node_wipe_cache(self, node: NodeInterface):
        node.wipe_cache()
        print(f"Wiped RID cache of '{node.name}'")
    
    @validate_args
    @load_node
    def node_wipe_logs(self, node: NodeInterface):
        node.wipe_logs()
        print(f"Wiped logs of '{node.name}'")
    
    @validate_args
    def network_sync(self):
        self.network.sync()
        
    @validate_args
    def network_wipe_cache(self):
        self.network.wipe_cache()
        
    @validate_args
    def network_wipe_logs(self):
        self.network.wipe_logs()
        
    @validate_args
    def network_wipe_config(self):
        self.network.wipe_config()
    
    @validate_args
    def network_status(self):
        table = Table("node", "state", box=box.SIMPLE)
        for node in self.network.nodes:
            node_state = node.state()
            match node_state:
                case NodeState.IDLE:
                    c = "white"
                case NodeState.STARTING:
                    c = "blue"
                case NodeState.RUNNING:
                    c = "green"
                case NodeState.STOPPING:
                    c = "red"
            
            table.add_row(node.name, f"[{c}]{node.state()}[/{c}]")
            
        self.console.print(table)
    
    @validate_args
    @load_node
    def network_set_first_contact(self, node: NodeInterface):
        self.network.set_first_contact(node)
    
    @validate_args
    def network_unset_first_contact(self):
        fc_node = self.network.get_first_contact()
        if fc_node:
            self.network.unset_first_contact(fc_node)
    
    @validate_args
    def network_run(self):
        self.network.run()
    
    @validate_args
    def network_start(self):
        self.network.start()
    
    @validate_args
    def network_stop(self):
        self.network.stop()
    
    @validate_args
    def module_list(self):
        table = Table("module", "alias(es)", box=box.SIMPLE)
        module_alias_map: dict[str, set[str]] = {}
        for alias, module in module_interface.alias_module_map.items():
            module_alias_map.setdefault(module, set()).add(alias)
        
        for module, alias_set in module_alias_map.items():
            table.add_row(module, ','.join(alias_set))
            
        self.console.print(table)
    
    @validate_args
    def module_reload(self, module_ref: str):
        try:
            module = module_interface.resolve_ref(module_ref)
        except ModuleNotFoundError:
            self.console.print(f"Couldn't resolve module reference '{module_ref}'")
        
        # can only reload modules of nodes at rest
        
        updated_node_class = module_interface.load_class(module, reload_module=True)
        
        affected_nodes = 0
        updated_nodes = 0
        for node in self.network.nodes:
            if node.module != module:
                continue
            
            affected_nodes += 1
            
            if node.state() != NodeState.IDLE:
                self.console.print(f"Skipping running node '{node.name}'")
                continue
            
            node.set_node_class(updated_node_class)
            updated_nodes += 1
        
        if affected_nodes == 0:
            self.console.print("No nodes were affected")
        else:
            self.console.print(f"Reload module for {updated_nodes}/{affected_nodes} nodes")
        
def run():
    KoiShell().cmdloop()


from pydantic import BaseModel, Field

from ..build.component import provides, CompType
from .env_config import EnvConfig
from .koi_net_config import KoiNetConfig


@provides(CompType.OBJECT)
class BaseNodeConfig(BaseModel):
    """Base node config class, intended to be extended.
    
    Using the `comp_type.object` decorator to mark this class as an
    object to be treated "as is" rather than attempting to initialize it
    during the build.
    """
    
    koi_net: KoiNetConfig
    # note: EnvConfig has to use a default factory, otherwise it will
    # evaluated during the library import and cause an error if any
    # env variables are undefined
    env: EnvConfig = Field(default_factory=EnvConfig)


"""Defines KOI-net protocol errors."""

from enum import StrEnum
from ..exceptions import (
    ProtocolError,
    UnknownNodeError,
    InvalidKeyError,
    InvalidSignatureError,
    InvalidTargetError
)


class ErrorType(StrEnum):
    UnknownNode = "unknown_node"
    InvalidKey = "invalid_key"
    InvalidSignature = "invalid_signature"
    InvalidTarget = "invalid_target"

EXCEPTION_TO_ERROR_TYPE: dict[ProtocolError, ErrorType] = {
    UnknownNodeError: ErrorType.UnknownNode,
    InvalidKeyError: ErrorType.InvalidKey,
    InvalidSignatureError: ErrorType.InvalidSignature,
    InvalidTargetError: ErrorType.InvalidTarget
}


import threading
from logging import Logger
from pathlib import Path

import structlog
from structlog.contextvars import bound_contextvars

from ..logging_context import LoggingContext
from .assembler import NodeAssembler


class BaseAssembly(NodeAssembler):
    root_dir: Path
    
    shutdown_signal: threading.Event = threading.Event

    log: Logger = lambda root_dir: structlog.stdlib.get_logger().bind(log_dir=root_dir)
    logging_context: LoggingContext = LoggingContext
    
    def __new__(cls, *args, root_dir: Path, **kwargs):
        cls.root_dir = root_dir
        with bound_contextvars(log_dir=root_dir):
            return super().__new__(cls, *args, **kwargs)



from dataclasses import dataclass
from logging import Logger
from rid_lib.ext import Cache
from rid_lib.types import KoiNetNode

from ..build.component import depends_on
from ..exceptions import RequestError
from ..network.graph import NetworkGraph
from ..network.request_handler import RequestHandler
from ..processor.kobj_queue import KobjQueue
from ..protocol.node import NodeProfile, NodeType


@dataclass
class SyncManager:
    """Handles state synchronization actions with other nodes."""
    
    log: Logger
    graph: NetworkGraph
    cache: Cache
    request_handler: RequestHandler
    kobj_queue: KobjQueue
    
    @depends_on("graph", "kobj_worker")
    def start(self):
        """Catches up with node providers on startup."""
        
        node_providers = self.graph.get_neighbors(
            direction="in",
            allowed_type=KoiNetNode
        )
        
        if not node_providers:
            return
        
        self.log.debug(f"Catching up with `orn:koi-net.node` providers: {node_providers}")
        self.catch_up_with(node_providers, [KoiNetNode])
    
    def catch_up_with(self, nodes, rid_types):
        """Catches up with the state of RID types within other nodes."""
    
        for node in nodes:
            node_bundle = self.cache.read(node)
            node_profile = node_bundle.validate_contents(NodeProfile)
            
            # can't catch up with partial nodes
            if node_profile.node_type != NodeType.FULL:
                continue
            
            try:
                payload = self.request_handler.fetch_manifests(
                    node, rid_types=rid_types)
            except RequestError:
                continue
            
            for manifest in payload.manifests:
                self.kobj_queue.push(
                    manifest=manifest,
                    source=node
                )


from dataclasses import dataclass
from enum import StrEnum
from typing import Callable
from rid_lib.core import RIDType

from ..protocol.event import EventType
from .knowledge_object import KnowledgeObject
from .context import HandlerContext


class StopChain:
    """Class for STOP_CHAIN sentinel returned by knowledge handlers."""
    pass

STOP_CHAIN = StopChain()


class HandlerType(StrEnum):
    """Types of handlers used in knowledge processing pipeline.
    
    - RID - provided RID; if event type is `FORGET`, this handler decides whether to delete the knowledge from the cache by setting the normalized event type to `FORGET`, otherwise this handler decides whether to validate the manifest (and fetch it if not provided).
    - Manifest - provided RID, manifest; decides whether to validate the bundle (and fetch it if not provided).
    - Bundle - provided RID, manifest, contents (bundle); decides whether to write knowledge to the cache by setting the normalized event type to `NEW` or `UPDATE`.
    - Network - provided RID, manifest, contents (bundle); decides which nodes (if any) to broadcast an event about this knowledge to. (Note, if event type is `FORGET`, the manifest and contents will be retrieved from the local cache, and indicate the last state of the knowledge before it was deleted.)
    - Final - provided RID, manifests, contents (bundle); final action taken after network broadcast.
    """
    
    RID = "rid", 
    Manifest = "manifest", 
    Bundle = "bundle", 
    Network = "network", 
    Final = "final"

@dataclass
class KnowledgeHandler:
    """Handles knowledge processing events of the provided types."""
    
    func: Callable[[HandlerContext, KnowledgeObject], None | KnowledgeObject | StopChain]
    handler_type: HandlerType
    rid_types: tuple[RIDType] = ()
    event_types: tuple[EventType | None] = ()
    
    def __call__(
        self, 
        ctx: HandlerContext, 
        kobj: KnowledgeObject
    ) -> None | KnowledgeObject | StopChain:
        return self.func(ctx, kobj)
    
    @classmethod
    def create(
        cls,
        handler_type: HandlerType,
        rid_types: tuple[RIDType] = (),
        event_types: tuple[EventType | None] = ()
    ):
        """Decorator wraps a function, returns a KnowledgeHandler."""
        def decorator(func: Callable) -> KnowledgeHandler:
            handler = cls(func, handler_type, rid_types, event_types)
            return handler
        return decorator



"""Pydantic models for request and response objects in the KOI-net API."""

from typing import Annotated, Literal
from pydantic import BaseModel, Field
from rid_lib import RID, RIDType
from rid_lib.ext import Bundle, Manifest
from .event import Event
from .errors import ErrorType


# REQUEST MODELS

class PollEvents(BaseModel):
    type: Literal["poll_events"] = Field("poll_events")
    limit: int = 0
    
class FetchRids(BaseModel):
    type: Literal["fetch_rids"] = Field("fetch_rids")
    rid_types: list[RIDType] = []
    
class FetchManifests(BaseModel):
    type: Literal["fetch_manifests"] = Field("fetch_manifests")
    rid_types: list[RIDType] = []
    rids: list[RID] = []
    
class FetchBundles(BaseModel):
    type: Literal["fetch_bundles"] = Field("fetch_bundles")
    rids: list[RID]
    

# RESPONSE/PAYLOAD MODELS

class RidsPayload(BaseModel):
    type: Literal["rids_payload"] = Field("rids_payload")
    rids: list[RID]

class ManifestsPayload(BaseModel):
    type: Literal["manifests_payload"] = Field("manifests_payload")
    manifests: list[Manifest]
    not_found: list[RID] = []
    
class BundlesPayload(BaseModel):
    type: Literal["bundles_payload"] = Field("bundles_payload")
    bundles: list[Bundle]
    not_found: list[RID] = []
    deferred: list[RID] = []
    
class EventsPayload(BaseModel):
    type: Literal["events_payload"] = Field("events_payload")
    events: list[Event]
    

# ERROR MODELS

class ErrorResponse(BaseModel):
    type: Literal["error_response"] = Field("error_response")
    error: ErrorType

# TYPES

type RequestModels = EventsPayload | PollEvents | FetchRids | FetchManifests | FetchBundles
type ResponseModels = RidsPayload | ManifestsPayload | BundlesPayload | EventsPayload | ErrorResponse

type ApiModels = Annotated[
    RequestModels | ResponseModels,
    Field(discriminator="type")
]


from rid_lib.types import KoiNetNode
import structlog
from base64 import b64decode, b64encode
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import serialization
from rid_lib.ext.utils import sha256_hash
from cryptography.hazmat.primitives.asymmetric.utils import (
    decode_dss_signature, 
    encode_dss_signature
)

log = structlog.stdlib.get_logger()


def der_to_raw_signature(der_signature: bytes, curve=ec.SECP256R1()) -> bytes:
    """Converts a DER-encoded signature to raw r||s format."""
    
    # Decode the DER signature to get r and s
    r, s = decode_dss_signature(der_signature)
    
    # Determine byte length based on curve bit size
    byte_length = (curve.key_size + 7) // 8
    
    # Convert r and s to big-endian byte arrays of fixed length
    r_bytes = r.to_bytes(byte_length, byteorder='big')
    s_bytes = s.to_bytes(byte_length, byteorder='big')
    
    # Concatenate r and s
    return r_bytes + s_bytes


def raw_to_der_signature(raw_signature: bytes, curve=ec.SECP256R1()) -> bytes:
    """Converts a raw r||s signature to DER format."""
    
    # Determine byte length based on curve bit size
    byte_length = (curve.key_size + 7) // 8
    
    # Split the raw signature into r and s components
    if len(raw_signature) != 2 * byte_length:
        raise ValueError(f"Raw signature must be {2 * byte_length} bytes for {curve.name}")
    
    r_bytes = raw_signature[:byte_length]
    s_bytes = raw_signature[byte_length:]
    
    # Convert bytes to integers
    r = int.from_bytes(r_bytes, byteorder='big')
    s = int.from_bytes(s_bytes, byteorder='big')
    
    # Encode as DER
    return encode_dss_signature(r, s)


class PrivateKey:
    priv_key: ec.EllipticCurvePrivateKey
    
    def __init__(self, priv_key):
        self.priv_key = priv_key
    
    @classmethod
    def generate(cls):
        """Generates a new `Private Key`."""
        return cls(priv_key=ec.generate_private_key(ec.SECP256R1()))

    def public_key(self) -> "PublicKey":
        """Returns instance of `PublicKey` dervied from this private key."""
        return PublicKey(self.priv_key.public_key())
    
    @classmethod
    def from_pem(cls, priv_key_pem: str, password: str):
        """Loads `PrivateKey` from encrypted PEM string."""
        return cls(
            priv_key=serialization.load_pem_private_key(
                data=priv_key_pem.encode(),
                password=password.encode()
            )
        )

    def to_pem(self, password: str) -> str:
        """Saves `PrivateKey` to encrypted PEM string."""
        return self.priv_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.BestAvailableEncryption(password.encode())
        ).decode()
        
    def sign(self, message: bytes) -> str:
        """Returns base64 encoded raw signature bytes of the form r||s."""
        hashed_message = sha256_hash(message.decode())
        
        der_signature_bytes = self.priv_key.sign(
            data=message,
            signature_algorithm=ec.ECDSA(hashes.SHA256())
        )
        
        raw_signature_bytes = der_to_raw_signature(der_signature_bytes)
        
        signature = b64encode(raw_signature_bytes).decode()
        
        log.debug(f"Signing message with [{self.public_key().to_der()}]")
        log.debug(f"hash: {hashed_message}")
        log.debug(f"signature: {signature}")
        
        return signature


class PublicKey:
    pub_key: ec.EllipticCurvePublicKey
    
    def __init__(self, pub_key):
        self.pub_key = pub_key
    
    @classmethod
    def from_pem(cls, pub_key_pem: str):
        """Loads `PublicKey` from PEM string."""
        return cls(
            pub_key=serialization.load_pem_public_key(
                data=pub_key_pem.encode()
            )
        )
        
    def to_pem(self) -> str:
        """Saves `PublicKey` to PEM string."""
        return self.pub_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        ).decode()
        
    @classmethod
    def from_der(cls, pub_key_der: str):
        """Loads `PublicKey` from base64 encoded DER string."""
        return cls(
            pub_key=serialization.load_der_public_key(
                data=b64decode(pub_key_der)
            )
        )
    
    def to_der(self) -> str:
        """Saves `PublicKey` to base64 encoded DER string."""
        return b64encode(
            self.pub_key.public_bytes(
                encoding=serialization.Encoding.DER,
                format=serialization.PublicFormat.SubjectPublicKeyInfo
            )
        ).decode()
        
    def to_node_rid(self, name) -> KoiNetNode:
        """Returns an orn:koi-net.node RID from hashed DER string."""
        return KoiNetNode(
            name=name,
            hash=sha256_hash(self.to_der())
        )
        
    def verify(self, signature: str, message: bytes):
        """Verifies a signature for a message.
        
        Raises `cryptography.exceptions.InvalidSignature` on failure.
        """
        
        raw_signature_bytes = b64decode(signature)
        der_signature_bytes = raw_to_der_signature(raw_signature_bytes)
        
        self.pub_key.verify(
            signature=der_signature_bytes,
            data=message,
            signature_algorithm=ec.ECDSA(hashes.SHA256())
        )



from enum import StrEnum


START_FUNC_NAME = "start"
STOP_FUNC_NAME = "stop"

COMPONENT_TYPE_FIELD = "__component_type__"
DEPENDS_ON_FIELD = "__depends_on__"

class CompType(StrEnum):
    SINGLETON = "SINGLETON"
    OBJECT = "OBJECT"

def provides(component_type: CompType):
    def decorator(obj):
        setattr(obj, COMPONENT_TYPE_FIELD, component_type)
        return obj
    return decorator

def depends_on(*components):
    def decorator(obj):
        setattr(obj, DEPENDS_ON_FIELD, set(components))
        return obj
    return decorator


import os
import sys
import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path
from datetime import datetime
from typing import Callable

import structlog
import colorama


shared_log_processors: list[Callable] = [
    structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    structlog.stdlib.PositionalArgumentsFormatter(),
    structlog.processors.TimeStamper(fmt="iso"),
    structlog.processors.UnicodeDecoder(),
    structlog.processors.CallsiteParameterAdder({
        structlog.processors.CallsiteParameter.MODULE,
        structlog.processors.CallsiteParameter.FUNC_NAME
    }),
    structlog.contextvars.merge_contextvars
]


class PartitionedFileHandler(logging.Handler):
    def __init__(
        self,
        log_file_name: str = "log.ndjson",
        max_log_file_size: int = 10 * 1024 ** 2,
        num_log_file_backups: int = 5,
        log_file_encoding: str = "utf-8"
    ):
        self.handlers: dict[str, RotatingFileHandler] = {}
        self.log_file_name = log_file_name
        self.max_log_file_size = max_log_file_size
        self.max_log_file_backups = num_log_file_backups
        self.log_file_encoding = log_file_encoding
        
        self.processor_formatter = structlog.stdlib.ProcessorFormatter(
            processor=structlog.processors.JSONRenderer(),
            foreign_pre_chain=shared_log_processors
        )
        
        self.dropped_log_handler = RotatingFileHandler(
            filename="dropped_logs.txt",
            maxBytes=self.max_log_file_size,
            backupCount=self.max_log_file_backups,
            encoding=self.log_file_encoding,
            delay=True
        )
        
        super().__init__()
        
    def del_handler(self, log_dir: str, wipe_logs: bool = False):
        if log_dir in self.handlers:
            self.handlers[log_dir].close()
            if wipe_logs:
                try:
                    os.remove(self.handlers[log_dir].baseFilename)
                except OSError:
                    pass
            del self.handlers[log_dir]
        
    def get_handler(self, log_dir: str):
        if log_dir not in self.handlers:
            file_handler = RotatingFileHandler(
                filename=Path(log_dir) / Path(self.log_file_name),
                maxBytes=self.max_log_file_size,
                backupCount=self.max_log_file_backups,
                encoding=self.log_file_encoding,
                delay=True
            )
            
            file_handler.setFormatter(self.processor_formatter)
            
            file_handler.setLevel(logging.DEBUG)
            self.handlers[log_dir] = file_handler
        
        return self.handlers[log_dir]
    
    def emit(self, record: logging.LogRecord):
        if record.log_dir is not None:
            log_dir = record.log_dir
            
        elif type(record.msg) is dict and "log_dir" in record.msg:
            log_dir = record.msg["log_dir"]
        
        else:
            self.dropped_log_handler.emit(record)
            return
        
        self.get_handler(str(log_dir)).emit(record)
        

class LogSystem:
    """Configures and initializes the logging system."""
    
    use_file_handler: bool
    use_console_handler: bool
    file_handler_log_level: int
    console_handler_log_level: int
    
    _instance = None
    
    def __new__(
        cls,
        use_file_handler: bool = True,
        use_console_handler: bool = True,
        file_handler_log_level: int = logging.DEBUG,
        console_handler_log_level: int = logging.DEBUG
    ):
        """Only instantiable once, other calls will return the first object."""
        if not cls._instance:
            obj = super().__new__(cls)
            obj.use_file_handler = use_file_handler
            obj.use_console_handler = use_console_handler
            obj.file_handler_log_level = file_handler_log_level
            obj.console_handler_log_level = console_handler_log_level
            
            obj.configure()
            cls._instance = obj
            
        return cls._instance
    
    @staticmethod
    def delete_file_handler(log_dir: str, wipe_logs: bool = False):
        for handler in logging.getLogger().handlers:
            if isinstance(handler, PartitionedFileHandler):
                handler.del_handler(log_dir, wipe_logs=wipe_logs)
        
    def configure(self):
        handlers = []
        if self.use_file_handler:
            handlers.append(PartitionedFileHandler())
        if self.use_console_handler:
            handlers.append(self.configure_console_handler())
        
        logging.basicConfig(level=logging.DEBUG, handlers=handlers)
        old_factory = logging.getLogRecordFactory()

        def record_factory(*args, **kwargs):
            record = old_factory(*args, *kwargs)
            
            ctx = structlog.contextvars.get_contextvars()
            record.log_dir = ctx.get("log_dir")
            
            return record
        
        logging.setLogRecordFactory(record_factory)
        
        structlog.configure(
            processors=shared_log_processors + [
                structlog.stdlib.ProcessorFormatter.wrap_for_formatter],
            wrapper_class=structlog.stdlib.BoundLogger,
            logger_factory=structlog.stdlib.LoggerFactory(),
            cache_logger_on_first_use=True,
        )
        
    def configure_console_handler(self):
        console_renderer = structlog.dev.ConsoleRenderer(
            columns=[
                # Render the timestamp without the key name in yellow.
                structlog.dev.Column(
                    "timestamp",
                    structlog.dev.KeyValueColumnFormatter(
                        key_style=None,
                        value_style=colorama.Style.DIM,
                        reset_style=colorama.Style.RESET_ALL,
                        value_repr=lambda t: datetime.fromisoformat(t).strftime("%Y-%m-%d %H:%M:%S"),
                    ),
                ),
                structlog.dev.Column(
                    "level",
                    structlog.dev.LogLevelColumnFormatter(
                        level_styles={
                            level: colorama.Style.BRIGHT + color
                            for level, color in {
                                "critical": colorama.Fore.RED,
                                "exception": colorama.Fore.RED,
                                "error": colorama.Fore.RED,
                                "warn": colorama.Fore.YELLOW,
                                "warning": colorama.Fore.YELLOW,
                                "info": colorama.Fore.GREEN,
                                "debug": colorama.Fore.GREEN,
                                "notset": colorama.Back.RED,
                            }.items()
                        },
                        reset_style=colorama.Style.RESET_ALL,
                        width=9
                    )
                ),
                # Render the event without the key name in bright magenta.
                
                # Default formatter for all keys not explicitly mentioned. The key is
                # cyan, the value is green.
                structlog.dev.Column(
                    "path",
                    structlog.dev.KeyValueColumnFormatter(
                        key_style=None,
                        value_style=colorama.Fore.MAGENTA,
                        reset_style=colorama.Style.RESET_ALL,
                        value_repr=str,
                        width=30
                    ),
                ),
                structlog.dev.Column(
                    "event",
                    structlog.dev.KeyValueColumnFormatter(
                        key_style=None,
                        value_style=colorama.Fore.WHITE,
                        reset_style=colorama.Style.RESET_ALL,
                        value_repr=str,
                        width=30
                    ),
                ),
                structlog.dev.Column(
                    "",
                    structlog.dev.KeyValueColumnFormatter(
                        key_style=colorama.Fore.BLUE,
                        value_style=colorama.Fore.GREEN,
                        reset_style=colorama.Style.RESET_ALL,
                        value_repr=str,
                    ),
                )
            ]
        )
        
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(
            structlog.stdlib.ProcessorFormatter(
                processor=console_renderer,
                foreign_pre_chain=shared_log_processors
            )
        )
        
        console_handler.setLevel(self.console_handler_log_level)
        return console_handler


import threading
from dataclasses import dataclass, field
from logging import Logger
from queue import Queue

from rid_lib.core import RID
from rid_lib.ext import Bundle, Manifest
from rid_lib.types import KoiNetNode

from ..protocol.event import Event, EventType
from .knowledge_object import KnowledgeObject


@dataclass
class KobjQueue:
    """Queue for knowledge objects entering the processing pipeline."""
    log: Logger
    shutdown_signal: threading.Event
    
    q: Queue[KnowledgeObject] = field(init=False, default_factory=Queue)
    
    def push(
        self, *,
        rid: RID | None = None,
        manifest: Manifest | None = None,
        bundle: Bundle | None = None,
        event: Event | None = None,
        kobj: KnowledgeObject | None = None,
        event_type: EventType | None = None,
        source: KoiNetNode | None = None
    ):
        """Pushes knowledge object to queue.
        
        Input may take the form of an RID, manifest, bundle, event, 
        or knowledge object (with an optional event type for RID, 
        manifest, or bundle objects). All objects will be normalized 
        to knowledge objects and queued.
        """
        
        if rid:
            _kobj = KnowledgeObject.from_rid(rid, event_type, source)
        elif manifest:
            _kobj = KnowledgeObject.from_manifest(manifest, event_type, source)
        elif bundle:
            _kobj = KnowledgeObject.from_bundle(bundle, event_type, source)
        elif event:
            _kobj = KnowledgeObject.from_event(event, source)
        elif kobj:
            _kobj = kobj
        else:
            raise ValueError("One of 'rid', 'manifest', 'bundle', 'event', or 'kobj' must be provided")
        
        self.q.put(_kobj)
        self.log.debug(f"Queued {_kobj!r}")
    
    def wait(self):
        """Safe join, prevents deadlock if `kobj_worker` fails."""
        while not self.shutdown_signal.wait(0.1):
            if self.q.unfinished_tasks == 0:
                return
        
        print("WAIT FAILED")
        raise RuntimeError("Shutdown while awaiting queue")


from dataclasses import dataclass, field
from logging import Logger
from pathlib import Path

import cryptography.exceptions
from rid_lib.ext import Bundle, Cache
from rid_lib.ext.utils import sha256_hash
from rid_lib.types import KoiNetNode

from .config.provider import ConfigProvider
from .identity import NodeIdentity
from .protocol.envelope import UnsignedEnvelope, SignedEnvelope
from .protocol.secure import PublicKey
from .protocol.api_models import ApiModels, EventsPayload
from .protocol.event import EventType
from .protocol.node import NodeProfile
from .protocol.secure import PrivateKey
from .exceptions import (
    UnknownNodeError,
    InvalidKeyError,
    InvalidSignatureError,
    InvalidTargetError
)
from .config.base import BaseNodeConfig


@dataclass
class SecureManager:
    """Subsystem handling secure protocol logic."""
    
    log: Logger
    identity: NodeIdentity
    cache: Cache
    config: ConfigProvider | BaseNodeConfig
    root_dir: Path
    
    priv_key: PrivateKey = field(init=False)
    
    def __post_init__(self):
        self.load_priv_key()
        
    @property
    def pem_path(self) -> Path:
        return self.root_dir / self.config.koi_net.private_key_pem_path
    
    def create_priv_key(self):
        self.priv_key = PrivateKey.generate()
        
        with open(self.pem_path, "w") as f:
            f.write(self.priv_key.to_pem(self.config.env.priv_key_password))
        self.log.debug("Generated new private key, no PEM file found")
        
        pub_key = self.priv_key.public_key()
        self.config.koi_net.node_rid = pub_key.to_node_rid(
            name=self.config.koi_net.node_name)
        
        if self.config.koi_net.node_profile.public_key != pub_key.to_der():
            if self.config.koi_net.node_profile.public_key:
                self.log.warning("New private key overwriting old public key!")
            
            self.config.koi_net.node_profile.public_key = pub_key.to_der()
        self.config.save_to_yaml()
    
    def load_priv_key(self):
        """Loads private key from PEM file path in config."""
        
        try:
            with open(self.pem_path, "r") as f:
                priv_key_pem = f.read()
                
                self.priv_key = PrivateKey.from_pem(
                    priv_key_pem=priv_key_pem,
                    password=self.config.env.priv_key_password
                )
        except FileNotFoundError:
            self.create_priv_key()
        
        except ValueError:
            self.log.error("Incorrect password, could not decrypt PEM")
            # TODO: figure out more graceful way of failing startup sequence
            raise
        
    def handle_unknown_node(self, envelope: SignedEnvelope) -> Bundle | None:
        """Attempts to find node profile in proided envelope.
        
        If an unknown node sends an envelope, it may still be able to be
        validated if that envelope contains their node profile. This is
        essential for allowing unknown nodes to handshake and introduce
        themselves. Only an `EventsPayload` contain a `NEW` event for a 
        node profile for the source node is permissible.
        """
        if type(envelope.payload) != EventsPayload:
            return None
            
        for event in envelope.payload.events:
            # must be NEW event for bundle of source node's profile
            if event.rid != envelope.source_node:
                continue
            if event.event_type != EventType.NEW:
                continue            
            
            return event.bundle
        return None
        
    def create_envelope(
        self, payload: ApiModels, target: KoiNetNode
    ) -> SignedEnvelope:
        """Returns signed envelope to target from provided payload."""
        return UnsignedEnvelope(
            payload=payload,
            source_node=self.identity.rid,
            target_node=target
        ).sign_with(self.priv_key)
        
    def validate_envelope(self, envelope: SignedEnvelope):
        """Validates signed envelope from another node."""
        
        node_bundle = (
            self.cache.read(envelope.source_node) or
            self.handle_unknown_node(envelope)
        )
        
        if not node_bundle:
            raise UnknownNodeError(f"Couldn't resolve {envelope.source_node}")
        
        node_profile = node_bundle.validate_contents(NodeProfile)
        
        # check that public key matches source node RID
        if envelope.source_node.hash != sha256_hash(node_profile.public_key):
            raise InvalidKeyError("Invalid public key on new node!")
        
        # check envelope signed by validated public key
        pub_key = PublicKey.from_der(node_profile.public_key)
        try:
            envelope.verify_with(pub_key)
        except cryptography.exceptions.InvalidSignature:
            raise InvalidSignatureError(f"Signature {envelope.signature} is invalid.")
        
        # check that this node is the target of the envelope
        if envelope.target_node != self.identity.rid:
            raise InvalidTargetError(f"Envelope target {envelope.target_node!r} is not me")




import queue
from dataclasses import dataclass

from ..config.base import BaseNodeConfig
from ..processor.pipeline import KnowledgePipeline
from ..processor.kobj_queue import KobjQueue
from ..build.threaded_component import ThreadedComponent
from ..build.component import depends_on


class End:
    """Class for STOP_WORKER sentinel pushed to worker queues."""
    pass

STOP_WORKER = End()


@dataclass
class KnowledgeProcessingWorker(ThreadedComponent):
    """Thread worker that processes the `kobj_queue`."""
    
    config: BaseNodeConfig
    kobj_queue: KobjQueue
    pipeline: KnowledgePipeline
    
    @depends_on("server", "poller")
    def stop(self):
        self.kobj_queue.q.put(STOP_WORKER)
        super().stop()
        
    def run(self):
        while True:
            try:
                item = self.kobj_queue.q.get(timeout=self.config.koi_net.kobj_worker.queue_timeout)
                try:
                    if item is STOP_WORKER:
                        self.log.info("Received 'STOP_WORKER' signal, shutting down...")
                        return
                    
                    self.log.info(f"Dequeued {item!r}")
                    self.pipeline.process(item)
                    
                finally:
                    self.kobj_queue.q.task_done()
                    
            except queue.Empty:
                pass




from enum import StrEnum
from pydantic import BaseModel
from rid_lib import RID
from rid_lib.ext import Manifest, Bundle


class EventType(StrEnum):
    NEW = "NEW"
    UPDATE = "UPDATE"
    FORGET = "FORGET"

class Event(BaseModel):
    rid: RID
    event_type: EventType
    manifest: Manifest | None = None
    contents: dict | None = None
    
    def __repr__(self):
        return f"<Event '{self.rid}' event type: '{self.event_type}'>"
    
    @classmethod
    def from_bundle(cls, event_type: EventType, bundle: Bundle):
        return cls(
            rid=bundle.manifest.rid,
            event_type=event_type,
            manifest=bundle.manifest,
            contents=bundle.contents
        )
        
    @classmethod
    def from_manifest(cls, event_type: EventType, manifest: Manifest):
        return cls(
            rid=manifest.rid,
            event_type=event_type,
            manifest=manifest
        )
        
    @classmethod
    def from_rid(cls, event_type: EventType, rid: RID):
        return cls(
            rid=rid,
            event_type=event_type
        )
        
    @property
    def bundle(self):
        if self.manifest is not None and self.contents is not None:
            return Bundle(
                manifest=self.manifest,
                contents=self.contents
            )


from queue import Queue

from rid_lib.types import KoiNetNode
from pydantic import BaseModel

from ..protocol.event import Event


class QueuedEvent(BaseModel):
    event: Event
    target: KoiNetNode

class EventQueue:
    """Queue for outgoing network events."""
    q: Queue[QueuedEvent]
    
    def __init__(self):
        self.q = Queue()
    
    def push(self, event: Event, target: KoiNetNode):
        """Pushes event to queue of specified node.
        
        Event will be sent to webhook or poll queue by the event worker
        depending on the node type and edge type of the specified node.
        """
        
        self.q.put(QueuedEvent(target=target, event=event))
    


from dataclasses import dataclass, field
from logging import Logger
from typing import Literal

import networkx as nx
from rid_lib import RIDType
from rid_lib.ext import Cache
from rid_lib.types import KoiNetEdge, KoiNetNode

from ..identity import NodeIdentity
from ..protocol.edge import EdgeProfile, EdgeStatus


@dataclass
class NetworkGraph:
    """Graph functions for this node's view of its network."""
    
    log: Logger
    cache: Cache
    identity: NodeIdentity
    
    dg: nx.DiGraph = field(init=False, default_factory=nx.DiGraph)
    
    def start(self):
        self.generate()
        
    def generate(self):
        """Generates directed graph from cached KOI nodes and edges."""
        self.log.debug("Generating network graph")
        self.dg.clear()
        for rid in self.cache.list_rids():
            if type(rid) == KoiNetNode:
                self.dg.add_node(rid)
                self.log.debug(f"Added node {rid!r}")
                
            elif type(rid) == KoiNetEdge:
                edge_bundle = self.cache.read(rid)
                if not edge_bundle:
                    self.log.warning(f"Failed to load {rid!r}")
                    continue
                edge_profile = edge_bundle.validate_contents(EdgeProfile)
                self.dg.add_edge(edge_profile.source, edge_profile.target, rid=rid)
                self.log.debug(f"Added edge {rid!r} ({edge_profile.source} -> {edge_profile.target})")
        self.log.debug("Done")
        
    def get_edge(
        self, 
        source: KoiNetNode, 
        target: KoiNetNode
    ) -> KoiNetEdge | None:
        """Returns edge RID given the RIDs of a source and target node."""
        if (source, target) in self.dg.edges:
            edge_data = self.dg.get_edge_data(source, target)
            if edge_data:
                return edge_data.get("rid")

        return None

    def get_edges(
        self,
        direction: Literal["in", "out"] | None = None,
    ) -> list[KoiNetEdge]:
        """Returns edges this node belongs to.
        
        All edges returned by default, specify `direction` to restrict 
        to incoming or outgoing edges only.
        """
        
        edges = []
        if (direction is None or direction == "out") and self.dg.out_edges:
            out_edges = self.dg.out_edges(self.identity.rid)
            edges.extend(out_edges)
        
        if (direction is None or direction == "in") and self.dg.in_edges:
            in_edges = self.dg.in_edges(self.identity.rid)
            edges.extend(in_edges)
        
        edge_rids = []
        for edge in edges:
            edge_data = self.dg.get_edge_data(*edge)
            if not edge_data: continue
            edge_rid = edge_data.get("rid")
            if not edge_rid: continue
            edge_rids.append(edge_rid)
       
        return edge_rids
    
    def get_neighbors(
        self,
        direction: Literal["in", "out"] | None = None,
        status: EdgeStatus | None = None,
        allowed_type: RIDType | None = None
    ) -> list[KoiNetNode]:
        """Returns neighboring nodes this node shares an edge with.
        
        All neighboring nodes returned by default, specify `direction` 
        to restrict to neighbors connected by incoming or outgoing edges
        only.
        """
        
        neighbors = set()
        for edge_rid in self.get_edges(direction):
            edge_bundle = self.cache.read(edge_rid)
            
            if not edge_bundle: 
                self.log.warning(f"Failed to find edge {edge_rid!r} in cache")
                continue
            
            edge_profile = edge_bundle.validate_contents(EdgeProfile)
                        
            if status and edge_profile.status != status:
                continue
            
            if allowed_type and allowed_type not in edge_profile.rid_types:
                continue
            
            if edge_profile.target == self.identity.rid:
                neighbors.add(edge_profile.source)
            elif edge_profile.source == self.identity.rid:
                neighbors.add(edge_profile.target)
            
        return list(neighbors)
